(** Conservative variant of the lisp collector. *)

(** mark_field proof *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Import cgc.
Require Import vc2.
Require Import talrules.
Require Import talgc.


(** ******************************************************************* *)
(** specifications and codes *)
(** ******************************************************************* *)

(** ******************************************************************* *)
(** return *)
(** ******************************************************************* *)
Definition return_pre (s:state) :=  True.
Definition return_guar (s s' :state) := s' = s.

Definition returnc := 
  jr ra.

Definition return1_pre (s:state) :=  True.
Definition return1_guar (s s' :state) := s' = rset s ra (rget s rcst1).

Definition return1 := 
  addiu ra rcst1 0;
  jr ra.


(** ******************************************************************* *)
(** mark stack *)
(** ******************************************************************* *)

(** is_empty() *)
Definition is_empty_pre s :=
  exists objs, s |||- set_hp objs s ** top.

Definition is_empty_guar s s' :=
  (forall objs A, s |||- set_hp objs s ** A ->
    sbool (rget s' v0) (objs = empty_set)) /\ 
  hp_of s = hp_of s' /\
  (forall r, r <> v0 -> rget s r = rget s' r).

Definition is_empty :=
  beq rbot rtop STACKEMPTYT;;
  addiu v0 zero 0;
  jr ra.

Definition is_emptyT_pre (s : state) := True.
Definition is_emptyT_guar (s s' : state):= s' = rset s v0 1.

Definition is_emptyT :=
  addiu v0 zero 1;
  jr ra.

(** push() *)
Definition push_pre s :=
  exists objs, ~ set_in (rget s a0) objs /\
    s |||- set_hp objs s ** top.

Definition push_guar s s' :=
  (forall objs A, s |||- set_hp objs s ** A -> 
    ~ set_in (rget s a0) objs ->
    s' |||- set_hp (set_add (rget s a0) objs) s' ** A) /\
  (forall r, r <> rtop -> r <> rtemp -> rget s r = rget s' r).

Definition push :=
  addiu rtemp rtop 4;
  sltu rtemp rbuff rtemp;
  bne rtemp zero LOOP;;
  sw a0 0 rtop;
  addiu rtop rtop 4;
  jr ra.

Definition loop_pre (s : state) := True.
Definition loop_guar (s s' : state):= False.

Definition loop :=
  j LOOP.

(** pop() *)
Definition pop_pre s :=
  exists objs, (objs <> empty_set) /\
    s |||- set_hp objs s ** top.

Definition pop_guar s s' :=
  (forall objs A, s |||- set_hp objs s ** A -> 
    objs <> empty_set ->
    set_in (rget s' v0) objs /\
    s' |||- set_hp (set_remove (rget s' v0) objs) s' ** A) /\
  (forall r, r <> v0 -> r <> rtop -> rget s r = rget s' r).

Definition pop :=
  addiu v0 zero 4;
  subu rtop rtop v0;
  lw v0 0 rtop;
  jr ra.

(* mark_field(val) { *)
(*   if (val < ohpSt) || (val >= ohpEd) *)
(*     return; *)
(*   if ((val - ohpSt) mod 8 <> 0) *)
(*     return; *)
(*   if (val->color == BLACK)  *)
(*     return; *)
(*   val->color = BLACK; *)
(*   push(val); *)
(* } *)

(** mark_field() *)
Definition mark_field_pre s :=
  (rget s rost = ost) /\ (rget s roed = oed) /\
  (~ vptr' (rget s a0) \/ (* not a valid pointer *)
    vptr' (rget s a0) /\
    exists G, 
      (set_in (rget s a0) G /\ 
	s |||- obj_hdrs G MARKED ** set_hp G s ** top) \/
      (~ set_in (rget s a0) G /\
	s |||- hdr_addr (rget s a0) |->_ ** obj_hdrs G MARKED ** set_hp G s ** top)).


Definition mark_field_guar s s' :=
  (((~ vptr' (rget s a0)) \/ 
    s |||- hdr_addr (rget s a0) |-> MARKED ** top) -> 
  hp_of s' = hp_of s /\ rget s' rtop = rget s rtop) /\
  (forall A G, 
    vptr' (rget s a0) -> 
    s |||- hdr_addr (rget s a0) |-> UNMARK ** obj_hdrs G MARKED ** set_hp G s ** A ->
    s' |||- obj_hdrs (set_add (rget s a0) G) MARKED **
      set_hp (set_add (rget s a0) G) s' ** A) /\
  reg_nin_eq (ast::rtemp2::rtop::v1::nil) s s'.

Definition mark_field :=
  sltu rtemp2 a0 rost;
  bne rtemp2 zero RETURN;; (* val < ohpst *)
  sltu rtemp2 a0 roed;
  beq rtemp2 zero RETURN;; (* ~ val < ohped *)
  subu ast a0 rost; (* ast = val - ohpst *)
  andi7 rtemp2 ast;
  bne rtemp2 zero RETURN;; (* (val - ohpst) mod 8 <> 0 *)
  srl1 ast ast;
  addu ast ast roed; (* ast = (val - ohpst)/2 + ohped *)
  lw rtemp2 0 ast;
  addiu v1 zero 1;
  beq rtemp2 v1 RETURN;; (* val->color = BLACK *)
  sw v1 0 ast;
  j STACKPUSH.
     
(** ******************************************************************* *)
(** mark *)
(** ******************************************************************* *)
Definition mark_pre_p s h:=
  sted_ok s /\
  s |||- hp_eq h ** obj_hdrs ohp_ptrs UNMARK ** set_hp empty_set s /\
  h ||- obj_hp ohp_ptrs ohp_ptrs.

Definition mark_post s h:= exists B, exists W,
  sted_ok s /\ roots_ok B s /\ set_union B W = ohp_ptrs /\
  s |||- hp_eq h ** obj_hdrs W UNMARK ** obj_hdrs B MARKED ** set_hp empty_set s /\
  h ||- obj_hp B B ** obj_hp ohp_ptrs W.

Definition mark_pre s := exists h, mark_pre_p s h.

Definition mark_guar s s' :=
  (forall h, mark_pre_p s h -> mark_post s' h) /\
  reg_nin_eq (rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'.

Definition mark :=
  addiu rcst1 ra 0;
  addiu a0 s1 0;
  jal MARK1 MARKFIELD.

Definition mark1_pre_p s h:= exists B, exists G, exists W,
  sted_ok s /\
  ok_val (set_union B G) (rget s s1) /\
  set_union (set_union B G) W = ohp_ptrs /\
  s |||- hp_eq h ** obj_hdrs B MARKED ** obj_hdrs G MARKED ** obj_hdrs W UNMARK ** set_hp G s /\
  h ||- obj_hp (set_union B G) B ** obj_hp ohp_ptrs (set_union G W).

Definition mark1_pre s := exists h, mark1_pre_p s h.

Definition mark1_guar s s' :=
  (forall h, mark1_pre_p s h -> mark_post s' h) /\
  reg_nin_eq (ra::rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'/\
  rget s' ra = rget s rcst1.

Definition mark1 :=
  addiu a0 s2 0;
  jal MARK2 MARKFIELD.

Definition mark2_pre_p s h:= exists B, exists G, exists W,
  sted_ok s /\
  ok_val (set_union B G) (rget s s1) /\
  ok_val (set_union B G) (rget s s2) /\
  set_union (set_union B G) W = ohp_ptrs /\
  s |||- hp_eq h ** obj_hdrs B MARKED ** obj_hdrs G MARKED ** obj_hdrs W UNMARK ** set_hp G s /\
  h ||- obj_hp (set_union B G) B ** obj_hp ohp_ptrs (set_union G W).

Definition mark2_pre s := exists h, mark2_pre_p s h.

Definition mark2_guar s s' :=
  (forall h, mark2_pre_p s h -> mark_post s' h) /\
  reg_nin_eq (ra::rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'/\
  rget s' ra = rget s rcst1.

Definition mark2 :=
  addiu a0 rcst3 0;
  jal MARKLOOP MARKFIELD.

Definition mark_loop_inv s h := exists B, exists G, exists W,
  sted_ok s /\
  roots_ok (set_union B G) s /\
  set_union (set_union B G) W = ohp_ptrs /\
  s |||- hp_eq h ** obj_hdrs B MARKED ** obj_hdrs G MARKED ** obj_hdrs W UNMARK ** set_hp G s /\
  h ||- obj_hp (set_union B G) B ** obj_hp ohp_ptrs (set_union G W).

Definition mark_loop_pre s := exists h, mark_loop_inv s h.

Definition mark_loop_guar s s' :=
  (forall h, mark_loop_inv s h -> mark_post s' h) /\
  reg_nin_eq (ra::rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'/\
  rget s' ra = rget s rcst1.

Definition mark_loop :=
  jal MARKLOOPBODY STACKEMPTY.

Definition mark_loop_body_p s h := exists B, exists G, exists W,
  sted_ok s /\
  sbool (rget s v0) (G = empty_set) /\ 
  roots_ok (set_union B G) s /\
  set_union (set_union B G) W = ohp_ptrs /\
  s |||- hp_eq h ** obj_hdrs B MARKED ** obj_hdrs G MARKED ** obj_hdrs W UNMARK ** set_hp G s /\
  h ||- obj_hp (set_union B G) B ** obj_hp ohp_ptrs (set_union G W).

Definition mark_loop_body_pre s := exists h, mark_loop_body_p s h.

Definition mark_loop_body_guar s s' :=
  (forall h, mark_loop_body_p s h -> mark_post s' h) /\
  reg_nin_eq (ra::rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'/\
  rget s' ra = rget s rcst1.

Definition mark_loop_body :=
  bne v0 zero RETURN1;;
  jal MARKFIRST STACKPOP.

Definition mark_first_pre_p s h := exists B, exists G, exists W,
  sted_ok s /\
  roots_ok (set_add (rget s v0) (set_union B G)) s /\ 
  set_union (set_add (rget s v0) (set_union B G)) W = ohp_ptrs /\
  s |||- hp_eq h ** hdr_addr (rget s v0) |-> MARKED **
    obj_hdrs B MARKED ** obj_hdrs G MARKED ** obj_hdrs W UNMARK ** set_hp G s /\
  h ||- obj_hp (set_add (rget s v0) (set_union B G)) B ** obj_hp ohp_ptrs (set_union G W) **
    ok_obj ohp_ptrs (rget s v0).

Definition mark_first_pre s :=
  exists h, mark_first_pre_p s h.

Definition mark_first_guar s s' :=
  (forall h, mark_first_pre_p s h -> mark_post s' h) /\
  reg_nin_eq (ra::rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'/\
  rget s' ra = rget s rcst1.

Definition mark_first :=
  lw a0 0 v0;
  jal MARKSECOND MARKFIELD.

Definition mark_second_pre_p s h := exists B, exists G, exists W,
  sted_ok s /\
  roots_ok (set_add (rget s v0) (set_union B G)) s /\ 
  set_union (set_add (rget s v0) (set_union B G)) W = ohp_ptrs /\
  s |||- hp_eq h ** hdr_addr (rget s v0) |-> MARKED **
    obj_hdrs B MARKED ** obj_hdrs G MARKED ** obj_hdrs W UNMARK ** set_hp G s /\
  h ||- obj_hp (set_add (rget s v0) (set_union B G)) B ** obj_hp ohp_ptrs (set_union G W) **
    ok_fld (set_add (rget s v0) (set_union B G)) (rget s v0) **
    ok_fld ohp_ptrs (rget s v0 + 4).

Definition mark_second_pre s :=
  exists h, mark_second_pre_p s h.

Definition mark_second_guar s s' :=
  (forall h, mark_second_pre_p s h -> mark_post s' h) /\
  reg_nin_eq (ra::rcst1::a0::ast::v0::rtemp2::rtop::v1::nil) s s'/\
  rget s' ra = rget s rcst1.

Definition mark_second :=
  lw a0 4 v0;
  jal MARKLOOP MARKFIELD.


(** ******************************************************************* *)
(** sweep *)
(** ******************************************************************* *)
Definition sweep_pre_p s h := exists B, exists W,
  sted_ok s /\ roots_ok B s /\ set_union B W = ohp_ptrs /\
  s |||- hp_eq h ** obj_hp ohp_ptrs W ** obj_hdrs W UNMARK ** 
    obj_hdrs B MARKED ** set_hp empty_set s /\
  h ||- obj_hp B B.

Definition sweep_pre s := exists h, sweep_pre_p s h.

Definition sweep_post s h := exists B, exists W,
  sted_ok s /\ roots_ok B s /\ set_union B W = ohp_ptrs /\
  s |||- hp_eq h ** flst_hp W s ** obj_hdrs ohp_ptrs UNMARK ** set_hp empty_set s /\
  h ||- obj_hp B B.

Definition sweep_guar s s' :=
  (forall h, sweep_pre_p s h -> sweep_post s' h) /\
  reg_nin_eq (rsweep::rfree::ast::v0::nil) s s'.

Definition sweep :=
  addiu rsweep rost 0;
  addiu rfree zero 0;
  j SWEEPLOOP.

Definition sweep_loop_inv s h := exists B, exists W, exists F, exists G,
  sted_ok s /\ roots_ok B s /\ G = ms_objs ost (S (rget s rsweep)) /\ 
  set_disj W G /\ set_subset F G /\ (exists k, rget s rsweep = ost + 8 * k) /\
  set_union (set_union B W) F = ohp_ptrs /\
  s |||- hp_eq h ** obj_hp ohp_ptrs W ** flst_hp F s ** 
    obj_hdrs W UNMARK ** obj_hdrs F UNMARK ** 
    bobj_hdrs B (rget s rsweep) UNMARK MARKED ** set_hp empty_set s /\
  h ||- obj_hp B B.

Definition sweep_loop_pre s := exists h, sweep_loop_inv s h.

Definition sweep_loop_guar s s' :=
  (forall h, sweep_loop_inv s h -> sweep_post s' h) /\
  reg_nin_eq (rsweep::rfree::ast::v0::nil) s s'.

Definition sweep_loop :=
  sltu ast rsweep roed;
  beq ast zero RETURN;; (* if rsweep >= roed return *)
  subu ast rsweep rost; (* ast = rsweep - ohpst *)
  srl1 ast ast;
  addu ast ast roed; (* ast = (rsweep - ohpst)/2 + ohped *)
  lw v0 0 ast; (* v0 = rsweep.markbit *)
  beq v0 zero SWEEPADD;; (* if white, collect *)
  sw zero 0 ast; (* else, reset *)
  j SWEEPNEXT.

Definition sweep_next_pre s :=
  sweep_loop_pre (rset s rsweep (rget s rsweep + 8)).

Definition sweep_next_guar s s' :=
  sweep_loop_guar (rset s rsweep (rget s rsweep + 8)) s'.

Definition sweep_add_pre s :=
  s |||- rget s rsweep + 4 |->_ ** top /\
  sweep_next_pre (rset (hset s (rget s rsweep + 4) (rget s rfree)) rfree (rget s rsweep)).

Definition sweep_add_guar s s' :=
  sweep_next_guar (rset (hset s (rget s rsweep + 4) (rget s rfree)) rfree (rget s rsweep)) s'.

Definition sweep_add :=
  sw rfree 4 rsweep;
  addiu rfree rsweep 0;
  j SWEEPNEXT.

Definition sweep_next :=
  addiu rsweep rsweep 8;
  j SWEEPLOOP.


(** ******************************************************************* *)
(** gc() *)
(** ******************************************************************* *)
Notation gc_pre_p := (fun s h => mark_pre_p (rset s rcst2 (rget s ra)) h).

Definition gc_pre := (fun s => exists h, gc_pre_p s h).

Definition gc_post s h := exists h',
  sweep_post s h' /\ h ||- hp_eq h' ** top.

Definition gc_guar s s' :=
  (forall h, gc_pre_p s h -> gc_post s' h) /\
  reg_nin_eq (rcst1::rcst2::rfree::a0::ast::v0::rtemp2::v1::nil) s s'.

Definition gc :=
  addiu rcst2 ra 0;
  jal GC2 MARK.

Definition gc2 :=
  addiu ra rcst2 0;
  j SWEEP.

Notation gc2_pre_p := (fun s h => sweep_pre_p (rset s ra (rget s rcst2)) h).

Definition gc2_pre s := exists h, gc2_pre_p s h.

Definition gc2_guar s s' := sweep_guar (rset s ra (rget s rcst2)) s' /\ 
  rget s' ra = rget s rcst2.


(** ******************************************************************* *)
(** alloc *)
(** ******************************************************************* *)

(* alloc(){ *)
(*   if (rfree == NULL) gc(); *)
(*   if (rfree == NULL) loop(); *)
(*   s2 = rfree; *)
(*   rfree = rfree -> next; *)
(* } *)

Definition alloc_pre s := exists h, gc_inv s h.

Definition alloc_guar s s' := forall h Psi G T1 T2,
  gc_inv s h ->
  WFwty T1 -> WFwty T2 ->
  WFst Psi G (rf_of s, h) -> 
  exists h', gc_inv s' h' /\
    WFst Psi (regty_set G s2 (tmem false T1 false T2)) (rf_of s', h').

Definition alloc :=
  addiu rcst3 ra 0;
  bne rfree zero ALLOCRETPTR;;
  jal ALLOCLOOP GC.

Notation ret_ptr_pre_p := (fun s x => hget s (rget s rfree + 4) = Some x).

Notation ret_ptr_pre := (fun s => exists x, ret_ptr_pre_p s x).

Notation ret_ptr_guar := (fun s s' => (forall x, 
  ret_ptr_pre_p s x -> s' = 
  rset (rset (rset s ra (rget s rcst3)) s2 (rget s rfree)) rfree x)).

Definition ret_ptr :=
  addiu ra rcst3 0;
  addiu s2 rfree 0;
  lw rfree 4 rfree;
  jr ra.

Definition alloc_loop_pre := (fun s =>
  rget s rfree <> 0 -> ret_ptr_pre s).
 
Definition alloc_loop_guar s s' := 
  rget s rfree <> 0 /\ ret_ptr_guar s s'.

Definition alloc_loop :=
  beq rfree zero LOOP;;
  j ALLOCRETPTR.


(** ******************************************************************* *)
(** gc Psi *)
(** ******************************************************************* *)
Definition gc_P :=
  (RETURN,,(isp2 return_pre return_guar)) ::: (
  (RETURN1,,(isp2 return1_pre return1_guar)) ::: (
  (LOOP,,(isp2 loop_pre loop_guar)):::(

  (STACKEMPTY,,(isp2 is_empty_pre is_empty_guar)) ::: (
  (STACKEMPTYT,,(isp2 is_emptyT_pre is_emptyT_guar)) ::: (
  (STACKPOP,,(isp2 pop_pre pop_guar)) ::: (
  (STACKPUSH,,(isp2 push_pre push_guar)) ::: (

  (MARKFIELD,,(isp2 mark_field_pre mark_field_guar)) ::: (

  (MARK,,(isp2 mark_pre mark_guar)) ::: (
  (MARK1,,(isp2 mark1_pre mark1_guar)) ::: (
  (MARK2,,(isp2 mark2_pre mark2_guar)) ::: (
  (MARKLOOP,,(isp2 mark_loop_pre mark_loop_guar)) ::: (
  (MARKLOOPBODY,,(isp2 mark_loop_body_pre mark_loop_body_guar)) ::: (
  (MARKFIRST,,(isp2 mark_first_pre mark_first_guar)) ::: (
  (MARKSECOND,,(isp2 mark_second_pre mark_second_guar)) ::: (

  (SWEEP,,(isp2 sweep_pre sweep_guar)):::(
  (SWEEPLOOP,,(isp2 sweep_loop_pre sweep_loop_guar)):::(
  (SWEEPADD,,(isp2 sweep_add_pre sweep_add_guar)):::(
  (SWEEPNEXT,,(isp2 sweep_next_pre sweep_next_guar)):::(

  (GC,,(isp2 gc_pre gc_guar)):::(
  (GC2,,(isp2 gc2_pre gc2_guar)):::(

  (ALLOC,,(isp2 alloc_pre alloc_guar)):::(
  (ALLOCLOOP,,(isp2 alloc_loop_pre alloc_loop_guar)):::(
  (ALLOCRETPTR,,(isp2 ret_ptr_pre ret_ptr_guar)):::(
  nilT _)))))))))))))))))))))))).


(** ******************************************************************* *)
(** return *)
(** ******************************************************************* *)
Theorem return_ok : 
  scap_iseq_ok gc_P return_pre return_guar returnc.
Proof.
  simplvc.
  unfold return_pre; unfold return_guar.
  firstorder.
Qed.

Theorem return1_ok : 
  scap_iseq_ok gc_P return1_pre return1_guar return1.
Proof.
  simplvc.
  unfold return1_pre; unfold return1_guar.
  repeat split; trivial; intros s'; simplpm; auto.
Qed.


Ltac unfoldpred :=
  unfold is_empty_pre; unfold is_empty_guar;
    unfold is_emptyT_pre; unfold is_emptyT_guar; 
      unfold push_pre; unfold push_guar; 
	unfold pop_pre; unfold pop_guar; 
	  unfold loop_pre; unfold loop_guar;
	    unfold set_hp; unfold buffer_stack_set;
	      unfold buffer_array_stack.

Ltac unfoldpred2 :=
  unfold mark_field_pre; unfold mark_field_guar;
    unfold push_pre; unfold push_guar. 


(** ******************************************************************* *)
(** is_empty() *)
(** ******************************************************************* *)
Theorem is_empty_ok : 
  scap_iseq_ok gc_P is_empty_pre is_empty_guar is_empty.
Proof.
  simplvc.
  unfoldpred.
  intros s [objs H].
  assert (Hold := H).
  sepelim H.
  simplif.
  assert (objs = empty_set).
  destruct A0; rewrite R in H0.
  assert (x = nil).
  destruct x; auto.
  simpl in * ; elimtype False; omega.
  subst.
  destruct A.
  hypreplace H2 (sd_eqv_nil_eq (sd_eqv_sym H2)).
  setsolver.
  unfold set_in in H4; rewrite H2 in H4; inversion H4.
  subst objs.
  rewrite R in Hold.
  split; trivial.

  clear Hold A H A0 x.
  intros; subst.
  split.
  intros.
  sepelim H.
  simplst.
  destruct A1.
  assert (x = nil).
  destruct x; auto.
  simpl in * ; elimtype False; omega.
  subst.
  destruct A0.
  hypreplace H2 (sd_eqv_nil_eq (sd_eqv_sym H2)).
  setsolver.
  unfold set_in in H4; rewrite H2 in H4; inversion H4.

  split;auto.
  intros; simplst; trivial.

  clear Hold A H A0 x objs.
  split;auto.

  intros; subst; simplst; intuition.
  sepelim H.
  subst objs; simpl in A0.
  omega.
  simplst; reflexivity.
Qed.

Theorem is_emptyT_ok : 
  scap_iseq_ok gc_P is_emptyT_pre is_emptyT_guar is_emptyT.
Proof.
  simplvc.
  unfoldpred.
  intuition.
Qed.

Theorem loop_ok : 
  scap_iseq_ok gc_P loop_pre loop_guar loop.
Proof.
  simplvc.
  unfoldpred.
  tauto.
Qed.

Theorem pop_ok : scap_iseq_ok gc_P pop_pre pop_guar pop.
Proof.
  simplvc.
  unfoldpred.

  intros s [objs [Hobjs Hpred]].
  sepelim Hpred.
  simplst.
  destruct x.
  elimtype False; apply Hobjs.
  destruct A.
  setsolver.
  hypreplace H (sd_eqv_nil_eq (sd_eqv_sym H)).
  unfold set_in in H1; rewrite H in H1; inversion H1.

  simpl in Hpred.
  simplpm; simpllw n 1.
  split; trivial.

  intros; subst; simplst.
  clear objs Hobjs A A0.
  split.
  intros.
  sepelim H.
  destruct x0.
  elimtype False; apply H0.
  destruct A0.
  setsolver.
  hypreplace H1 (sd_eqv_nil_eq (sd_eqv_sym H1)).
  unfold set_in in H3; rewrite H1 in H3; inversion H3.
  simpl in H.
  hvinject 1 1.
  subst n0; clear Hpred.
  split.
  destruct A0.
  destruct (H1 n).
  apply H4; left; reflexivity.
  lexintro 1; exists x0.
  simpl in * ; intuition.
  simplsep 4 4; try split.
  split; intros.
  destruct (sd_remove_elim H5).
  destruct (H1 n0).
  destruct (H8 H7); try tauto.
  assert (n0 <> n).
  intro F; subst n0.
  assert (sd_eqv (sd_of objs) x0).
  apply sd_eqv_trans with (n :: x0); trivial.
  split; intros.
  destruct H6; subst; auto.
  right; trivial.
  hypreplace H6 (sd_eqv_length_le H6).
  omega.
  apply sd_remove_intro2; trivial.
  destruct (H1 n0).
  apply H8; right; trivial.
  assert (set_in n objs).
  destruct (H1 n).
  apply H6; left; reflexivity.
  rewrite <- (set_remove_length_dec H5) in H2.
  simpl in H2.
  omega.
  omega.
  omega.
  apply buffer_add_head with n; trivial.
  omega.
  intros r He1 He2.
  simplst; reflexivity.
Qed.
    
Theorem push_ok : 
  scap_iseq_ok gc_P push_pre push_guar push.
Proof.
  simplvc.
  unfoldpred.

  intros s [objs [Ha0 Hpred]].
  sepelim Hpred.
  destruct A; destruct A0.
  simplst.
  simplif.
  sconjlifth 1 Hpred.
  assert (rget s rtop + 4 <= rget s rbuff).
  generalize R.
  simplcmp (rget s rbuff) (rget s rtop + 4); intro; try discriminate.
  rewrite (beq_nat_eq _ _ (sym_eq RE)); trivial.
  generalize (blt_lt RE); omega.
  hypreplace Hpred (buffer_remove_head H3 Hpred).
  sepelim Hpred.
  simplpm; simpllw x0 1.
  split; trivial.
  intros; subst s'.
  simplst.
  split; intros.
  sepelim H4.
  sconjlifth 1 H4.
  hypreplace H4 (buffer_remove_head H3 H4).
  sepelim H4.
  simpl.
  hvinject 1 1; subst; clear Hpred.
  simplpm.
  lexintro 1; exists ((rget s a0):: x1).
  hypreplace H4 (cont_update1 H4 (rget s a0)).
  destruct A0; destruct A1.
  simplsep 4 4; try split; try omega.
  
  split; intros; destruct (H6 n).
  destruct (sd_insert_elim H10); subst.
  left;reflexivity.
  right; auto.
  destruct H10; subst.
  apply sd_insert_intro1.
  apply sd_insert_intro2; auto.
  generalize (set_add_length_inc H5).
  simpl; omega.
  simpl.
  omega.
  simpl.
  simplsep 2 2.
  simplst; reflexivity.

  intuition.
Qed.

(** ******************************************************************* *)
(** mark_field() *)
(** ******************************************************************* *)
Ltac returnpf :=
  let r := fresh "r" in let F := fresh "F" in
    let s' := fresh "s'" in
      (split; [unfold return_pre; apply I |
	unfold return_guar; intro s'; intros; subst s';
	  repeat split; intros r F; try tauto; 
	    destruct F; intuition; simplst; reflexivity]).

Lemma mark_field_ok : 
  scap_iseq_ok gc_P mark_field_pre mark_field_guar mark_field.
Proof.
  simplvc.
  unfoldpred2.
  intros; simplst.

  destruct H as [St [Ed [Atom | [NAtom [G HG]]]]].

(*   case 1. atom value *)
  simplif.
  simplif.
  returnpf.

  simplif.
  elimtype False; apply Atom.
  generalize R R0; clear R R0.
  repeat simplif; intros; try discriminate.
  destruct (mod8_elim R1).
  apply ms_objs_intro with x.
  destruct (mod8_elim ost_mod8).
  destruct (mod8_elim oed_mod8).
  omega.
  omega.
  tauto.
  
  returnpf.
  returnpf.

(*   case 2: not atom *)
  destruct (ms_objs_elim NAtom) as [Ge [k Hk]].
  simplif.
  simplif.
  elimtype False; generalize R R0.
  repeat simplif; intros; try discriminate.
  omega.

  simplif.
  destruct HG as [[In Hp] | [NIn Hp]].

  hypreplace Hp (fas_elim In Hp).
  rewrite Plus.plus_comm.
  rewrite St; rewrite Ed.
  rewrite Plus.plus_comm; simplpm; rewrite Plus.plus_comm; simpllw 1 4.
  rewrite beq_true.

  split.  
  unfold return_pre; apply I.
  unfold return_guar; intros; subst s'.
  repeat split; intros. 
  elimtype False.
  simpl in Hp.
  hvinject 1 1; discriminate.
  destruct H; intuition; simplst; reflexivity. 

  sepelim Hp.
  rewrite Plus.plus_comm.
  rewrite St; rewrite Ed.
  rewrite Plus.plus_comm; simplpm; rewrite Plus.plus_comm; simpllw x 1.
  simplif.

  split.  
  unfold return_pre; apply I.
  unfold return_guar; intros; subst s'.
  repeat split; intros. 
  elimtype False.
  simpl in Hp.
  hvinject 1 1; subst; discriminate.
  destruct H; intuition; simplst; reflexivity. 

  simplpm; simpllw x 1.
  split.
  exists G.
  split; trivial.
  hypreplace Hp (cont_update1 Hp 1).
  simplsep 3 3.

  intros s' [Hg Hr].
  split; intros.
  clear Hg Hr.
  destruct H.
  tauto.
  hvinject 1 1; subst; tauto.
  split; [clear Hr | clear Hg].
  intros A G' Hptr Hp0.
  sconjlift 1.
  assert (~ set_in (rget s a0) G').
  intro F.
  sconjlifth 1 Hp0.
  hypreplace Hp0 (fas_elim F Hp0).
  simpl in Hp0; hvinject 1 1; subst; tauto.

  apply Hg; clear Hg; auto.
  hypreplace Hp0 (cont_update1 Hp0 1).
  simplsep 3 3.
  unfold obj_hdrs in * .
  apply fas_intro1; auto.

  intros.
  destruct H as [A [B [C [D E]]]].
  rewrite <- Hr; auto.
  simplst. reflexivity.
  
  replace (rget s a0 - rget s rost) with (8 * k) in R1.
  rewrite mod8_mul_8 in R1; tauto.
  omega.

  elimtype False.
  generalize R; simplif; intro; try tauto.
  omega.
Qed.



(** ******************************************************************* *)
(** mark() *)
(** ******************************************************************* *)
Lemma mark_field_guar_ra_ok : forall s s', 
  mark_field_guar s s' -> rget s ra = rget s' ra.
Proof.
  intros s s' [H [H' H'']].
  rewrite <- H''.
  reflexivity.
  repeat split; discriminate.
Qed.

Theorem mark_loop_body_ok : 
  scap_iseq_ok gc_P mark_loop_body_pre mark_loop_body_guar mark_loop_body.
Proof.
  simplvc.

  intros s Pre.
  simplif; simplst.

  apply call_chain.

  split.

  destruct Pre as [h [B [G [W [[Hst Hed] 
    [HG [Hroot [Hatom [Hobjs [Hp Hh]]]]]]]]]].
  rewrite R in HG; simpl in HG.

  exists G.
  split;auto.
  simplsep 2 7.

  intros s' [Hg Hr].
  assert (XXX : forall h,
    mark_loop_body_p s h -> mark_first_pre_p s' h).
  intros h Hh.
  destruct Hh as [B [G [W [[Hst Hed] 
    [HG [Hroot [Hobjs [Hp Hh]]]]]]]].
  rewrite R in HG; simpl in HG.

  sconjlifth 4 Hp.
  destruct (Hg _ _ Hp HG) as [Hv0' Hp'].
  exists B; exists (set_remove (rget s' v0) G); exists W.
  rewrite <- (Hr rost); try discriminate.
  rewrite <- (Hr roed); try discriminate.
  simplst.
  repeat split; trivial.
  intros r RR.
  rewrite <- (Hr r).
  rewrite rset_rget_neq.
  intro In; hypreplace Hroot (Hroot _ RR In).
  clear In; simplset.
  intro F; subst r; destruct RR as [X | [X | [X | X]]]; 
    try discriminate; trivial.
  intro F; subst r; destruct RR as [X | [X | [X | X]]]; 
    try discriminate; trivial.
  intro F; subst r; destruct RR as [X | [X | [X | X]]]; 
    try discriminate; trivial.

  rewrite <- Hobjs; apply set_ext; split; intros; simplset.
  unfold set_hp in Hp'.
  simplsep 7 7.
  sconjlift 1.
  unfold obj_hdrs in *; apply fas_elim1 with 
    (P := fun x => hdr_addr x |-> 1); auto.
  eapply sconj_mono;intros.
  apply Hh.
  unfold obj_hp in * .
  eapply fas_impl.
  apply H.
  unfold ok_obj; unfold ok_fld; intros.
  sepelim H0.
  lexintro 1; exists x.
  lexintro 3; exists x0.
  simplsep 4 4.
  intro F; hypreplace A (A F); clear F; simplset.
  intro F; hypreplace A0 (A0 F); clear F; simplset.
  unfold obj_hp in * .
  assert (set_in (rget s' v0) (set_union G W)).
  simplset.
  hypreplace H (fas_elim1 H0 H).
  simplsep 2 2.
  replace (set_union (set_remove (rget s' v0) G) W) 
    with (set_remove (rget s' v0) (set_union G W)).
  trivial.
  apply set_ext; split; intros; simplset.
  destruct (set_remove_elim H1).
  tauto.
  destruct (set_union_elim H1).
  destruct (set_remove_elim H2).
  tauto.
  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim Hv0' Hp).
  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim H2 Hp).
  elimtype False; apply cont_cont_false 
    with (hp_of s) (hdr_addr (rget s' v0)) 1 0 top.
  simplsep 3 7.

  split.
  destruct Pre as [h Pre].
  exists h.
  apply (XXX _ Pre).
  intros s'' [Hg' [Hcc Hra]]. 
  repeat split; auto.

  intros r Nin.
  rewrite (Hcc _ Nin).
  rewrite <- (Hr r).
  destruct Nin; simplst; reflexivity.
  destruct Nin; intuition.
  destruct Nin; intuition.
  rewrite Hra; rewrite <- Hr; try discriminate; reflexivity.

  unfold return1_guar; unfold return1_pre; 
    unfold mark_loop_body_guar.
  split; trivial.
  intros.
  subst s'; simplst.
  repeat split.
  intros.
  destruct H as [B [G [W [[Hst Hed] 
    [HG [Hroot [Hobjs [Hp Hh]]]]]]]].
  destruct (rget s v0).
  tauto.
  subst G.
  exists B; exists W.
  simplst.
  repeat (split; trivial).
  intros r In; hypreplace Hroot (Hroot _ In).
  rewrite rset_rget_neq.
  intro F; hypreplace Hroot (Hroot F).
  simplset.
  intro F; subst r.
  destruct In as [X |[X |[X | X]]]; try discriminate; trivial.
  rewrite <- Hobjs; apply set_ext; split; intros; simplset.
  simplsep 4 4.
  hypreplace Hp (unit_elim_right Hp).
  generalize Hp.
  unfold set_hp.
  unfold buffer_stack_set in Hp.
  unfold buffer_array_stack in Hp.
  sepelim Hp.
  intuition.
  rewrite set_union_empty_right in Hh.
  rewrite set_union_empty_left in Hh; trivial.

  intros r Nin.
  rewrite rset_rget_neq; trivial.
  destruct Nin; trivial.

  intros s s' [H H'].
  rewrite <- H'; trivial; discriminate.
Qed.

Theorem mark_loop_ok : 
  scap_iseq_ok gc_P mark_loop_pre mark_loop_guar mark_loop.
Proof.
  simplvc.

  intros s Pre.
  apply call_chain.

  split.
  destruct Pre as [h [B [G [W [[Hst Hed] [Hroot [Hatom [Hobjs [Hp Hh]]]]]]]]].
  exists G.
  simplsep 2 7.

  intros s' [Hg [Hs Hr]].
  assert (XXX : forall h, mark_loop_inv s h -> mark_loop_body_p s' h).
  intros h [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]].

  exists B; exists G; exists W.
  rewrite <- Hr; [idtac | discriminate].
  rewrite <- Hr; [simplst | discriminate].
  repeat split; trivial.
  apply Hg with top.
  simplsep 2 7.

  eapply roots_pred_next; eauto.
  simpl; intuition; subst; rewrite <- Hr; trivial; discriminate.

  rewrite <- Hs.
  simplsep 4 4.
  unfold set_hp in * .
  repeat (rewrite <- Hr; try discriminate).
  apply Hp.

  clear Hg; split.
  destruct Pre as [h P].
  exists h; auto.
  intros s'' [Hg [R1 R2]].
  repeat split; intros; auto.
  rewrite R1; trivial.
  destruct H as [A [B [C [D [E [F [G _]]]]]]].
  rewrite <- Hr; [ simplst; reflexivity | regin].
  rewrite <- Hr in R2; [apply R2 | regin].

  intros; destruct H as [_ [_ H]]; apply H; discriminate.
Qed.



Theorem mark_first_ok : 
  scap_iseq_ok gc_P mark_first_pre mark_first_guar mark_first.
Proof.
  simplvc.

  intros s Pre; simplpm.
  assert (H := Pre).
  destruct H as [h [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]]].

  unfold ok_obj in Hh; unfold ok_fld in Hh; sepelim Hh.
  assert (Hall := sconj_insert Hp Hh).
  assert (X : hget s (rget s v0) = Some x);
    [simpllw x 4; reflexivity | rewrite X].
  apply call_chain.
  
  split.
  repeat (split; simplst; trivial).
  destruct (set_in_dec x ohp_ptrs) as [In | Nin];
    [right | left]; trivial.
  split; trivial.

  exists G.
  destruct (set_in_dec x G) as [In2 | Nin2];
    [left | right]; split; trivial.
  simplsep 3 7.
  
  assert (In2 : set_in x (set_union (set_add (rget s v0) B) W)).
  rewrite <- Hobjs in In.
  simplset; tauto.
  destruct (set_union_elim In2) as [InBv0 | InW].
  destruct (set_add_elim InBv0) as [Eqv0 | InB].
  lexintro 1; exists 1.
  subst x; simplsep 5 5.

  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim InB Hp).
  lexintro 1; exists 1.
  simplsep 4 4.

  sconjlifth 4 Hp.
  hypreplace Hp (fas_elim InW Hp).
  lexintro 1; exists 0.
  simplsep 5 5.
  
  unfold mark_field_guar; simplst.
  intros s' [Hga [Hgna Hr]].
  clear h B G W Hst Hed Hroot Hobjs Hp A x0 A0 Hh Hall.
  assert (XXX : forall h, mark_first_pre_p s h -> mark_second_pre_p s' h).
  intros h [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]].
  
  unfold mark_second_pre_p.
  repeat (rewrite Hr; [idtac | regin]).
  simplst.
  unfold ok_obj in Hh; unfold ok_fld in Hh; sepelim Hh.
  assert (Hall := sconj_insert Hp Hh).
  assert (X1 : hget s (rget s v0) = Some x0);
    [simpllw x0 4; reflexivity | rewrite X1 in X; 
      injection X; intros; subst x0; clear X].
  
  assert (Case : (set_in x ohp_ptrs /\ set_in x W) \/ 
    (set_in x (set_add (rget s v0) (set_union B G)) \/ ~ set_in x ohp_ptrs)).
  destruct (set_in_dec x ohp_ptrs) as [In | Nin].
  destruct (set_in_dec x W) as [InW | NinW].
  left; tauto.
  right; left.
  rewrite <- Hobjs in In; simplset; tauto.
  right; right; trivial.
  destruct Case as [[In InW] | Case]; [clear Hga | clear Hgna].

  exists B; exists (set_add x G); 
    exists (set_remove x W).
  repeat split; trivial.
  assert (T : roots_pred (fun v : nat =>
    ok_val (set_add (rget s v0) (set_union B G)) v) s').
  eapply roots_pred_next; eauto.
  simpl; intuition; subst; rewrite Hr; regin.
  intros r T2 F; hypreplace F (T r T2 F); simplset.
  rewrite <- Hobjs; setsolver.
  assert (s' |||- obj_hdrs (set_add x G) 1 **
    set_hp (set_add x G) s' ** hp_eq h ** obj_hdrs B 1 ** 
    oed + div2 (rget s v0 - ost) |-> 1 ** obj_hdrs (set_remove x W) 0).
  apply Hgna; auto.

  sconjlifth 4 Hp. 
  hypreplace Hp (fas_elim InW Hp).
  simplsep 6 6.
  simplsep 6 6.
  replace (set_union (set_add x G)
    (set_remove x W))
  with (set_union G W); [trivial | setsolver].
  unfold ok_obj; unfold ok_fld.
  lexintro 3; exists x.
  lexintro 3; exists x1.
  simplsep 4 4.
  intro F; simplset.
  eapply obj_hp_subset; eauto.
  red; intros; simplset.

  exists B; exists G; exists W.
  destruct Hga as [Hhp Hrtop].
  destruct Case as [Inv0BG | Nin]; [right | left; trivial].
  destruct (set_add_elim Inv0BG) as [Inv0 | InBG].
  subst x; simplsep 2 8.
  destruct (set_union_elim InBG) as [InB | InG].
  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim InB Hp); simplsep 2 2.
  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim InG Hp); simplsep 2 2.
  repeat split; trivial.
  eapply roots_pred_next; eauto.
  simpl; intuition; subst; rewrite Hr; regin.
  rewrite Hhp; simplsep 4 4.
  unfold set_hp in * .
  rewrite Hrtop.
  rewrite (Hr rbot); [idtac | regin].
  rewrite (Hr rbuff); [trivial | regin].
  unfold ok_obj; unfold ok_fld.
  lexintro 3; exists x.
  lexintro 3; exists x1.
  simplsep 4 4.
  intro F; destruct Case as [InBG | Nin]; trivial; tauto.

  clear Hga Hgna; split.
  destruct Pre as [h P].
  exists h; auto.
  intros s'' [Hg [R1 R2]].
  repeat split; intros; auto.
  rewrite R1; trivial.
  destruct H as [A [B [C [D [E [F [G [H _]]]]]]]].
  rewrite Hr;[simplst; reflexivity | regin].
  rewrite Hr in R2; [ apply R2 | regin].

  apply mark_field_guar_ra_ok.
Qed.

  
Theorem mark_second_ok : 
  scap_iseq_ok gc_P mark_second_pre mark_second_guar mark_second.
Proof.
  simplvc.

  intros s Pre; simplpm.
  assert (H := Pre).
  destruct H as [h [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]]].

  unfold ok_obj in Hh; unfold ok_fld in Hh; sepelim Hh.
  assert (Hall := sconj_insert Hp Hh).
  assert (X : hget s (rget s v0 + 4)  = Some x0);
    [simpllw x0 4; reflexivity | rewrite X].
  apply call_chain.
  
  split.
  repeat (split; simplst; trivial).
  destruct (set_in_dec x0 ohp_ptrs) as [In | Nin];
    [right | left]; trivial.
  split; trivial.

  exists G.
  destruct (set_in_dec x0 G) as [In2 | Nin2];
    [left | right]; split; trivial.
  simplsep 3 7.
  
  assert (In2 : set_in x0 (set_union (set_add (rget s v0) B) W)).
  rewrite <- Hobjs in In.
  simplset; tauto.
  destruct (set_union_elim In2) as [InBv0 | InW].
  destruct (set_add_elim InBv0) as [Eqv0 | InB].
  lexintro 1; exists 1.
  subst x0; simplsep 5 5.

  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim InB Hp).
  lexintro 1; exists 1.
  simplsep 4 4.

  sconjlifth 4 Hp.
  hypreplace Hp (fas_elim InW Hp).
  lexintro 1; exists 0.
  simplsep 5 5.
  
  unfold mark_field_guar; simplst.
  intros s' [Hga [Hgna Hr]].
  clear h B G W Hst Hed Hroot Hobjs Hp A x A0 Hh Hall. 
  assert (XXX : forall h, mark_second_pre_p s h -> mark_loop_inv s' h).
  intros h [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]].
  
  unfold mark_loop_inv.
  repeat (rewrite Hr; [idtac | regin]).
  simplst.
  unfold ok_obj in Hh; unfold ok_fld in Hh; sepelim Hh.
  assert (Hall := sconj_insert Hp Hh).
  assert (X1 : hget s (rget s v0 + 4) = Some x1);
    [simpllw x1 4; reflexivity | rewrite X1 in X; 
      injection X; intros; subst x1; clear X].
  
  assert (Hv0B : ~ set_in (rget s v0) B).
  intro F.
  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim F Hp).
  subst; apply cont_cont_false with (hp_of s) (hdr_addr (rget s v0)) 1 1 top.
  simplsep 3 7.

  assert (Case : (set_in x0 ohp_ptrs /\ set_in x0 W) \/ 
    (set_in x0 (set_add (rget s v0) (set_union B G)) \/ ~ set_in x0 ohp_ptrs)).
  destruct (set_in_dec x0 ohp_ptrs) as [In | Nin].
  destruct (set_in_dec x0 W) as [InW | NinW].
  left; tauto.
  right; left.
  rewrite <- Hobjs in In; simplset; tauto.
  right; right; trivial.
  destruct Case as [[In InW] | Case]; [clear Hga | clear Hgna].

  exists (set_add (rget s v0) B); exists (set_add x0 G); 
    exists (set_remove x0 W).
  repeat split; trivial.
  assert (T : roots_pred (fun v : nat =>
    ok_val (set_add (rget s v0) (set_union B G)) v) s').
  eapply roots_pred_next; eauto.
  simpl; intuition; subst; rewrite Hr; regin.
  intros r T2 F; hypreplace F (T r T2 F); simplset.
  rewrite <- Hobjs; setsolver.
  assert (s' |||- obj_hdrs (set_add x0 G) 1 **
    set_hp (set_add x0 G) s' ** hp_eq h ** obj_hdrs B 1 ** 
    oed + div2 (rget s v0 - ost) |-> 1 ** obj_hdrs (set_remove x0 W) 0).
  apply Hgna; auto.

  sconjlifth 4 Hp. 
  hypreplace Hp (fas_elim InW Hp).
  simplsep 6 6.
  simplsep 6 6.
  unfold obj_hdrs; apply fas_intro1; auto.
  simplsep 4 4.

  replace (set_union (set_add x0 G)
    (set_remove x0 W))
  with (set_union G W); [trivial | setsolver].
  simplsep 3 7.
  unfold obj_hp in * ; apply fas_intro1; auto.
  hypreplace Hh (sconj_assoc_right Hh).
  eapply sconj_mono; eauto; intros.
  unfold ok_obj; unfold ok_fld.
  lexintro 3; exists x.
  lexintro 3; exists x0.
  simplsep 4 4.
  intro F; hypreplace A (A F); simplset.
  intro F; simplset.
  assert (XX := obj_hp_subset); unfold obj_hp in XX.
  eapply XX; eauto.
  red; intros; simplset.

  exists (set_add (rget s v0) B); exists G; exists W.
  destruct Hga as [Hhp Hrtop].
  destruct Case as [Inv0BG | Nin]; [right | left; trivial].
  destruct (set_add_elim Inv0BG) as [Inv0 | InBG].
  subst x0; simplsep 2 8.
  destruct (set_union_elim InBG) as [InB | InG].
  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim InB Hp); simplsep 2 2.
  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim InG Hp); simplsep 2 2.
  repeat split; trivial.
  assert (T : roots_pred (fun v : nat =>
    ok_val (set_add (rget s v0) (set_union B G)) v) s').
  eapply roots_pred_next; eauto.
  simpl; intuition; subst; rewrite Hr; regin.
  intros r T2 F; hypreplace F (T r T2 F); simplset.
  rewrite <- Hobjs; setsolver.
  rewrite Hhp.
  unfold set_hp in * .
  rewrite Hrtop.
  rewrite (Hr rbot); [idtac | regin].
  rewrite (Hr rbuff); [trivial | regin].
  simplsep 3 3.
  unfold obj_hdrs; apply fas_intro1; auto.
  simplsep 2 2.

  unfold obj_hp in * ; apply fas_intro1; auto.
  hypreplace Hh (sconj_assoc_right Hh).
  eapply sconj_mono; eauto; intros.
  unfold ok_obj; unfold ok_fld.
  lexintro 3; exists x.
  lexintro 3; exists x0.
  simplsep 4 4.
  intro F; hypreplace A (A F); simplset.
  intro F; destruct Case as [InBG | Nin]; simplset; tauto.
  assert (XX := obj_hp_subset); unfold obj_hp in XX.
  eapply XX; eauto.
  red; intros; simplset.

  clear Hga Hgna; split.
  destruct Pre as [h P].
  exists h; auto.
  intros s'' [Hg [R1 R2]].
  repeat split; intros; auto.
  rewrite R1; trivial.
  destruct H as [A [B [C [D [E [F [G [H _]]]]]]]].
  rewrite Hr;[simplst; reflexivity | regin].
  rewrite Hr in R2; [ apply R2 | regin].

  apply mark_field_guar_ra_ok.
Qed.

Theorem mark_ok : scap_iseq_ok gc_P mark_pre mark_guar mark.
Proof.
  simplvc.

  intros s Pre.
  simplpm; simplst.
  apply call_chain.
  split.

  destruct Pre as [h [[Hst Hed] [Hp Hh]]].

  repeat (split; simplst; trivial).
  destruct (set_in_dec (rget s s1) ohp_ptrs) as [In | Nin];
    [right | left]; trivial.
  split; trivial.

  exists empty_set; right.
  split.
  intro F; inversion F.
  lexintro 1; exists 0.
  unfold set_hp in * ; simplst.
  sconjlifth 1 Hp.
  hypreplace Hp (fas_elim In Hp).
  simplsep 4 7.
  apply unit_intro_left; simplsep 1 1.

  unfold mark_field_guar; simplst.
  intros s' [Hga [Hgna Hr]].
  assert (XXX : forall h, mark_pre_p s h -> mark1_pre_p s' h).
  intros h H.

  destruct H as [[Hst Hed] [Hp Hh]].
  unfold mark1_pre_p.
  rewrite (Hr rost); [idtac | regin].
  rewrite (Hr roed); [idtac | regin].
  rewrite (Hr s1); [simplst | regin].

  destruct (set_in_dec (rget s s1) ohp_ptrs) as [In | Nin].

  exists empty_set; exists (set_add (rget s s1) empty_set); 
    exists (set_remove (rget s s1) ohp_ptrs).
  regin.
  intros F; simplset.
  rewrite set_union_empty_left.
  setsolver.
  assert (s' |||- obj_hdrs (set_add (rget s s1) empty_set) 1 **
    set_hp (set_add (rget s s1) empty_set) s' ** hp_eq h **
    obj_hdrs (set_remove (rget s s1) ohp_ptrs) 0).
  apply Hgna; auto.

  sconjlifth 1 Hp. 
  hypreplace Hp (fas_elim In Hp).
  sconjlift 1; apply unit_intro_left.
  simplsep 5 5.
  sconjlift 1; apply unit_intro_left.
  simplsep 7 7.
  apply unit_intro_left.
  replace (set_union (set_add (rget s s1) empty_set)
    (set_remove (rget s s1) ohp_ptrs))
  with ohp_ptrs; [trivial | apply set_ext; split; intros; simplset].

  destruct Hga as [Hp' Hrtop]; auto.
  clear Hgna.

  exists empty_set; exists empty_set; exists ohp_ptrs.
  repeat split; trivial.
  setsolver.
  sconjlift 1; apply unit_intro_left.
  apply sconj_assoc_right.
  apply unit_intro_left.
  rewrite Hp'.
  simplsep 3 3.
  generalize Hp; unfold set_hp.
  rewrite (Hr rbot); [idtac | regin].
  rewrite (Hr rbuff); [rewrite Hrtop | regin].
  simplst; trivial.

  repeat (rewrite set_union_empty_left).
  apply unit_intro_left; trivial.

  split.
  destruct Pre as [h Pre].
  exists h; auto.

  intros s'' [Hg [Hcc Hra]]. 
  split.

  intros.
  apply Hg; auto.

  intros r RR. 
  destruct RR as [A [B [C [D [E [F [G _]]]]]]].
  comparer r ra;
  [rewrite Hra | rewrite Hcc; [idtac | regin]].
  rewrite Hr;[simplst; reflexivity | regin].
  rewrite Hr;[simplst; reflexivity | regin].

  intros s s' [_ [_ H]]; rewrite <- H; regin.
Qed.

Theorem mark2_ok : scap_iseq_ok gc_P mark2_pre mark2_guar mark2.
Proof.
  simplvc.

  intros s Pre.
  simplpm; simplst.
  apply call_chain.
  split.

  destruct Pre as [h [B [G [W [[Hst Hed] 
    [Hs1 [Hs2 [Hobjs [Hp Hh]]]]]]]]].

  repeat (split; simplst; trivial).
  destruct (set_in_dec (rget s rcst3) ohp_ptrs) as [In | Nin];
    [right | left]; trivial.
  split; trivial.

  exists G.
  destruct (set_in_dec (rget s rcst3) G) as [In2 | Nin2];
    [left | right]; split; trivial.
  simplsep 3 3.
  
  assert (In2 : set_in (rget s rcst3) (set_union B W)).
  rewrite <- Hobjs in In.
  simplset; tauto.
  destruct (set_union_elim In2) as [InB | InW].
  
  sconjlifth 1 Hp.
  hypreplace Hp (fas_elim InB Hp).
  lexintro 1; exists 1.
  simplsep 4 4.

  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim InW Hp).
  lexintro 1; exists 0.
  simplsep 4 4.
  
  unfold mark_field_guar; simplst.
  intros s' [Hga [Hgna Hr]].
  assert (XXX : forall h, mark2_pre_p s h -> mark_loop_inv s' h).
  intros h H.

  destruct H as [B [G [W [[Hst Hed] [Hs1 [Hs2 [Hobjs [Hp Hh]]]]]]]].

  unfold mark_loop_inv.
  rewrite (Hr rost); [idtac | regin].
  rewrite (Hr roed); [idtac | regin].
  repeat (rewrite rset_rget_neq; try discriminate).
  simplst.

  assert (Case : (set_in (rget s rcst3) ohp_ptrs /\ set_in (rget s rcst3) W) \/ 
    (set_in (rget s rcst3) (set_union B G) \/ ~ set_in (rget s rcst3) ohp_ptrs)).
  destruct (set_in_dec (rget s rcst3) ohp_ptrs) as [In | Nin].
  destruct (set_in_dec (rget s rcst3) W) as [InW | NinW].
  left; tauto.
  right; left.
  rewrite <- Hobjs in In; simplset; tauto.
  right; right; trivial.
  destruct Case as [[In InW] | Case]; [clear Hga | clear Hgna].

  exists B; exists (set_add (rget s rcst3) G); 
    exists (set_remove (rget s rcst3) W).
  regin.
  intros r R.
  destruct R as [R | [R | [R | F]]]; subst.
  rewrite Hr; [simplst; intro F; hypreplace F (Hs1 F); simplset | regin].
  rewrite Hr; [simplst; intro F; hypreplace F (Hs2 F); simplset | regin].
  rewrite Hr; [simplst; intro F; simplset | regin].
  tauto.
  rewrite <- Hobjs; setsolver.
  assert (s' |||- obj_hdrs (set_add (rget s rcst3) G) 1 **
    set_hp (set_add (rget s rcst3) G) s' ** hp_eq h ** obj_hdrs B 1 **
    obj_hdrs (set_remove (rget s rcst3) W) 0).
  apply Hgna; auto.

  sconjlifth 3 Hp. 
  hypreplace Hp (fas_elim InW Hp).
  simplsep 5 5.
  simplsep 5 5.
  replace (set_union (set_add (rget s rcst3) G)
    (set_remove (rget s rcst3) W))
  with (set_union G W); [trivial | setsolver].
  simplsep 2 2.
  eapply obj_hp_subset; eauto.
  red; intros; simplset.

  exists B; exists G; exists W.
  destruct Hga as [Hhp Hrtop].
  destruct Case as [InBG | Nin]; [right | left; trivial].
  destruct (set_union_elim InBG) as [InB | InG].
  sconjlifth 1 Hp.
  hypreplace Hp (fas_elim InB Hp); simplsep 2 2.
  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim InG Hp); simplsep 2 2.
  repeat split; trivial.
  intros r R; destruct R as [R | [R | [R | F]]]; subst.
  rewrite Hr; [trivial | regin].
  rewrite Hr; [trivial | regin].
  destruct Case as [InBG | Nin]; 
    (rewrite Hr; [simplst; intro F | regin]).
  rewrite <- Hobjs in F; simplset.
  tauto.
  tauto.

  rewrite Hhp; simplsep 4 4.
  unfold set_hp in * .
  rewrite Hrtop.
  rewrite (Hr rbot); [idtac | regin].
  rewrite (Hr rbuff); [trivial | regin].

  clear Hga Hgna; split.
  destruct Pre as [h P].
  exists h; auto.
  intros s'' [Hg [R1 R2]].
  repeat split; intros; auto.
  rewrite R1; trivial.
  destruct H as [A [B [C [D [E [F [G [H _]]]]]]]].
  rewrite Hr;[simplst; reflexivity | regin].
  rewrite Hr in R2; [ apply R2 | regin].

  apply mark_field_guar_ra_ok.
Qed.

Theorem mark1_ok : scap_iseq_ok gc_P mark1_pre mark1_guar mark1.
Proof.
  simplvc.

  intros s Pre.
  simplpm; simplst.
  apply call_chain.
  split.

  destruct Pre as [h [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]]].

  repeat (split; simplst; trivial).
  destruct (set_in_dec (rget s s2) ohp_ptrs) as [In | Nin];
    [right | left]; trivial.
  split; trivial.

  exists G.
  destruct (set_in_dec (rget s s2) G) as [In2 | Nin2];
    [left | right]; split; trivial.
  simplsep 3 3.
  
  assert (In2 : set_in (rget s s2) (set_union B W)).
  rewrite <- Hobjs in In.
  simplset; tauto.
  destruct (set_union_elim In2) as [InB | InW].
  
  sconjlifth 1 Hp.
  hypreplace Hp (fas_elim InB Hp).
  lexintro 1; exists 1.
  simplsep 4 4.

  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim InW Hp).
  lexintro 1; exists 0.
  simplsep 4 4.
  
  unfold mark_field_guar; simplst.
  intros s' [Hga [Hgna Hr]].
  assert (XXX : forall h, mark1_pre_p s h -> mark2_pre_p s' h).
  intros h H.

  destruct H as [B [G [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]]].

  unfold mark2_pre_p.
  rewrite (Hr rost); [idtac | regin].
  rewrite (Hr roed); [idtac | regin].
  repeat (rewrite rset_rget_neq; try discriminate).
  rewrite (Hr s1); [idtac | regin].
  rewrite (Hr s2); [idtac | regin].
  simplst.

  assert (Case : (set_in (rget s s2) ohp_ptrs /\ set_in (rget s s2) W) \/ 
    (set_in (rget s s2) (set_union B G) \/ ~ set_in (rget s s2) ohp_ptrs)).
  destruct (set_in_dec (rget s s2) ohp_ptrs) as [In | Nin].
  destruct (set_in_dec (rget s s2) W) as [InW | NinW].
  left; tauto.
  right; left.
  rewrite <- Hobjs in In; simplset; tauto.
  right; right; trivial.
  destruct Case as [[In InW] | Case]; [clear Hga | clear Hgna].

  exists B; exists (set_add (rget s s2) G); 
    exists (set_remove (rget s s2) W).
  regin.
  intros F; hypreplace Hroot (Hroot F); simplset.
  intros F; simplset.
  rewrite <- Hobjs; setsolver.
  assert (s' |||- obj_hdrs (set_add (rget s s2) G) 1 **
    set_hp (set_add (rget s s2) G) s' ** hp_eq h ** obj_hdrs B 1 **
    obj_hdrs (set_remove (rget s s2) W) 0).
  apply Hgna; auto.

  sconjlifth 3 Hp. 
  hypreplace Hp (fas_elim InW Hp).
  simplsep 5 5.
  simplsep 5 5.
  replace (set_union (set_add (rget s s2) G)
    (set_remove (rget s s2) W))
  with (set_union G W); [trivial | setsolver].
  simplsep 2 2.
  eapply obj_hp_subset; eauto.
  red; intros; simplset.

  exists B; exists G; exists W.
  destruct Hga as [Hhp Hrtop].
  destruct Case as [InBG | Nin]; [right | left; trivial].
  destruct (set_union_elim InBG) as [InB | InG].
  sconjlifth 1 Hp.
  hypreplace Hp (fas_elim InB Hp); simplsep 2 2.
  sconjlifth 2 Hp.
  hypreplace Hp (fas_elim InG Hp); simplsep 2 2.
  regin.
  intro F; destruct Case as [InBG | Nin]; trivial; tauto.
  rewrite Hhp; simplsep 4 4.
  unfold set_hp in * .
  rewrite Hrtop.
  rewrite (Hr rbot); [idtac | regin].
  rewrite (Hr rbuff); [trivial | regin].

  clear Hga Hgna; split.
  destruct Pre as [h P].
  exists h; auto.
  intros s'' [Hg [R1 R2]].
  repeat split; intros; auto.
  rewrite R1; trivial.
  destruct H as [A [B [C [D [E [F [G [H _]]]]]]]].
  rewrite Hr; [ simplst; reflexivity | regin].
  rewrite Hr in R2; [apply R2 | regin].

  apply mark_field_guar_ra_ok.
Qed.


(** ******************************************************************* *)
(** sweep() *)
(** ******************************************************************* *)
Theorem sweep_ok : 
  scap_iseq_ok gc_P sweep_pre sweep_guar sweep.
Proof.
  simplvc.
  intros s Pre; simplpm.
  assert (XXX : forall h, sweep_pre_p s h ->
    sweep_loop_inv (rset (rset s rsweep (rget s rost)) rfree 0) h).
  intros h [B [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]].
  exists B; exists W; simplst; exists empty_set;
    exists (ms_objs ost (S (rget s rost))).
  repeat split; trivial.
  eapply roots_pred_next; eauto.
  intros r R.
  destruct R as [R |[R |[R | R]]]; subst; try tauto.
  intros x F1 F2.
  destruct (ms_objs_elim F2) as [Ge [k Hk]].
  rewrite <- Hst in Hk; omega.
  red; intros.
  inversion_clear H.
  exists 0; omega.
  rewrite <- Hobjs; setsolver.
  simplsep 7 5.
  unfold free_list.
  lexintro 1; exists (nil (A := nat)).
  simpl; simplsep 4 1.
  split; trivial.
  apply sd_eqv_refl.
  apply unit_intro_left.
  apply bobj_hdrs_intro; trivial.
  intros.
  hypreplace H (set_union_intro1 W H).
  rewrite Hobjs in H.
  destruct (ms_objs_elim H) as [Ge [k Hk]].
  rewrite Hst; omega.

  split.
  destruct Pre as [h H]; exists h; auto.

  intros s' [Hg Hr].
  split; intros; auto.
  rewrite (Hr _ H).
  destruct H as [A [B [C [D _]]]].
  simplst; reflexivity.
Qed.

Theorem sweep_next_ok : 
  scap_iseq_ok gc_P sweep_next_pre sweep_next_guar sweep_next.
Proof.
  simplvc.
  tauto.
Qed.

Theorem sweep_add_ok : 
  scap_iseq_ok gc_P sweep_add_pre sweep_add_guar sweep_add.
Proof.
  simplvc.
  intros s [Hp H].
  sepelim Hp.
  simpllw x 1.
  simplpm; tauto.
Qed.

Theorem sweep_loop_ok : 
  scap_iseq_ok gc_P sweep_loop_pre sweep_loop_guar sweep_loop.
Proof.
  simplvc.

  intros s Pre.

  unfold return_guar; unfold return_pre.
  simplif; generalize R; simplst; clear R; simplif; intro.
  discriminate.

  split; trivial.
  intros s' Hg; subst s'.
  split; intros.
  destruct H as [B [W [F [G [[Hst Hed] [Hroot 
    [HG [HW [HF [[k Hk] [Hobjs [Hp Hh]]]]]]]]]]]].

  assert (GG : set_subset (set_union (set_union B W) F) G).
  rewrite Hobjs; subst G.
  apply ms_objs_subset.
  lwomega.
  assert (WW : W = empty_set).
  apply empty_set_intro; intros x X.
  apply (HW _ X).
  apply GG; simplset.
  subst W; clear HW.
  rewrite set_union_empty_right in Hobjs.
  exists B; exists F.
  simplst; repeat split; trivial.
  eapply roots_pred_next; eauto.
  intros r RR.
  destruct RR as [X |[X |[ X | X]]]; subst; try reflexivity; tauto.
  simplsep 5 5.
  repeat (hypreplace Hp (unit_elim_left Hp)).
  rewrite <- Hobjs.
  apply obj_hdrs_merge; simplsep 2 2.
  apply bobj_hdrs_elim with 1 (rget s rsweep); trivial.
  intros x X.
  assert (XX : set_in x ohp_ptrs).
  rewrite <- Hobjs; simplset.
  destruct (ms_objs_elim XX) as [A1 [xx A2]].
  lwomega.
  destruct H as [A [B [C [D _]]]].
  simplst; reflexivity.

  assert (H := Pre).
  destruct H as [h [B [W [F [G [[Hst Hed] [Hroot 
    [HG [HW [HF [[k Hk] [Hobjs [Hp Hh]]]]]]]]]]]]].
  repeat rewrite Hst; repeat rewrite Hed; simplpm.

  assert (hget s (hdr_addr (rget s rsweep)) = Some 1 \/
    hget s (hdr_addr (rget s rsweep)) = Some 0).
  assert (set_in (rget s rsweep) (set_union B W)).
  assert (set_in (rget s rsweep) ohp_ptrs).
  apply ms_objs_intro with k; trivial.
  rewrite Hed in R. 
  unfold oed in * ; unfold ost in * ; lwomega.
  rewrite <- Hobjs in H; simplset.
  subst G; destruct (ms_objs_elim (HF _ SS)) as [FF _].
  elimtype False; lwomega.
  destruct (set_union_elim H) as [InB | InW]; [left | right].
  sconjlifth 5 Hp.
  generalize (fas_elim InB Hp).
  simplif; [elimtype False; lwomega | clear Hp; intro].
  simpllw 1 1; reflexivity.
  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim InW Hp).
  simpllw 0 1; reflexivity.

  rewrite Plus.plus_comm; simplpm.
  clear B W F G h Hst Hed Hroot HG HW HF k Hk Hobjs Hp Hh.
  destruct H as [Case1 | Case1]; rewrite Case1; simplst; simplbeq.
  simplpm; rewrite Case1.
  unfold sweep_next_pre; unfold sweep_next_guar; simplst.
  assert (XXX : forall h, sweep_loop_inv s h ->
    sweep_loop_inv (rset(hset(rset(rset
      (rset (rset (rset s rtemp 1) rtemp (rget s rsweep - ost))
        rtemp (div2 (rget s rsweep - ost))) rtemp
      (oed + div2 (rget s rsweep - ost))) v0 1)
    (oed + div2 (rget s rsweep - ost)) 0) rsweep (rget s rsweep + 8)) h).
  intros h [B [W [F [G [[Hst Hed] [Hroot 
    [HG [HW [HF [[k Hk] [Hobjs [Hp Hh]]]]]]]]]]]].

  exists B; exists W; exists F; 
    exists (ms_objs ost (S (rget s rsweep + 8))); simplst.
  assert (Hsweep : ~ set_in (rget s rsweep) W).
  intro InW.
  assert (hget s (hdr_addr (rget s rsweep)) = Some 0).
  sconjlifth 3 Hp.
  hypreplace Hp (fas_elim InW Hp).
  simpllw 0 1; reflexivity.
  rewrite H in Case1; discriminate.
  repeat split; trivial.
  eapply roots_pred_next; eauto.
  intros r RR; destruct RR as [X |[X |[X | X]]]; subst; try reflexivity; tauto.
  intros x X1 X2.
  apply (HW x X1).
  destruct (ms_objs_elim X2) as [A1 [xx A2]].
  rewrite HG; apply ms_objs_intro with xx; trivial.
  compare (rget s rsweep) x; intro e.
  rewrite <- e in X1; tauto.
  lwomega.
  eapply set_subset_trans; eauto.
  subst G; apply ms_objs_subset; lwomega.
  exists (S k); lwomega.
  assert (InB : set_in (rget s rsweep) B).
  assert (set_in (rget s rsweep) (set_union B W)).
  assert (set_in (rget s rsweep) ohp_ptrs).
  apply ms_objs_intro with k; trivial.
  rewrite Hed in R. 
  unfold oed in * ; unfold ost in * ; lwomega.
  rewrite <- Hobjs in H; simplset.
  subst G; destruct (ms_objs_elim (HF _ SS)) as [FF _].
  elimtype False; lwomega.
  destruct (set_union_elim H) as [InB | InW]; [trivial | tauto].
  sconjlifth 5 Hp.
  generalize (fas_elim InB Hp).
  simplif; [elimtype False; lwomega | clear Hp; intro Hp].
  hypreplace Hp (cont_update1 Hp 0). 
  simplsep 5 5.
  replace B with (set_add (rget s rsweep) (set_remove (rget s rsweep) B)).
  unfold bobj_hdrs; apply fas_intro1.
  simplif.
  simplsep 2 2.
  eapply fas_impl1; intros.
  apply Hp.
  replace (if blt_nat a (rget s rsweep + 8) then 0 else 1)
    with (if blt_nat a (rget s rsweep) then 0 else 1); trivial.
  assert (A : set_in a ohp_ptrs).
  rewrite <- Hobjs; simplset.
  destruct (ms_objs_elim A) as [X1 [kk Hl]].
  destruct (set_remove_elim H).
  simplif; simplif; try reflexivity;
    elimtype False; omega.
  elimtype False; omega.
  intro XX; destruct (set_remove_elim XX); tauto.
  setsolver.

  split.
  destruct Pre as [h Pre].
  exists h; auto.
  intros s' [Hg Hr]; split; intros; auto.
  rewrite Hr; trivial.
  destruct H as [A [B [C [D _]]]]; simplst; reflexivity.

  unfold sweep_add_pre; unfold sweep_add_guar; 
    unfold sweep_next_pre; unfold sweep_next_guar; simplst.
  assert (XXX : forall h, sweep_loop_inv s h ->
    sweep_loop_inv (rset (rset (hset (rset (rset (rset
      (rset (rset s rtemp 1) rtemp (rget s rsweep - ost))
      rtemp (div2 (rget s rsweep - ost))) rtemp
    (oed + div2 (rget s rsweep - ost))) v0 0)
    (rget s rsweep + 4) (rget s rfree)) rfree 
    (rget s rsweep)) rsweep (rget s rsweep + 8)) h).
  intros h [B [W [F [G [[Hst Hed] [Hroot 
    [HG [HW [HF [[k Hk] [Hobjs [Hp Hh]]]]]]]]]]]].

  exists B; exists (set_remove (rget s rsweep) W); 
    exists (set_add (rget s rsweep) F); exists (ms_objs ost (S (rget s rsweep + 8))).
  assert (NinB : ~ set_in (rget s rsweep) B).
  intro InB; assert (hget s (hdr_addr (rget s rsweep)) = Some 1).
  sconjlifth 5 Hp.
  generalize (fas_elim InB Hp).
  simplif; [elimtype False; lwomega | clear Hp; intro Hp].
  simpllw 1 1; reflexivity.
  rewrite H in Case1; discriminate.
  assert (NinF : ~ set_in (rget s rsweep) F).
  intro FF.
  hypreplace FF (HF _ FF).
  rewrite HG in FF; destruct (ms_objs_elim FF).
  lwomega.

  simplst; repeat split; trivial.
  eapply roots_pred_next; eauto.
  intros r RR; destruct RR as [X |[X |[X | X]]]; subst; try reflexivity; tauto.
  intros x X1 X2.
  destruct (set_remove_elim X1) as [X3 X4].
  apply (HW x X4).
  destruct (ms_objs_elim X2) as [A1 [xx A2]].
  rewrite HG; apply ms_objs_intro with xx; trivial.
  lwomega.
  intros x Hx.
  destruct (set_add_elim Hx) as [E | InF].
  subst x; apply ms_objs_intro with k; trivial.
  lwomega.
  hypreplace Hx (HF _ InF).
  rewrite HG in Hx; destruct (ms_objs_elim Hx) as [A1 [xx A2]].
  apply ms_objs_intro with xx; trivial.
  hypreplace InF (HF _ InF).
  rewrite HG in InF; destruct (ms_objs_elim InF).
  lwomega.
  exists (S k); lwomega.
  rewrite <- Hobjs; setsolver.
  rewrite Hobjs; apply ms_objs_intro with k; trivial.
  rewrite Hed in R.
  unfold ost in * ; unfold oed in * ;lwomega.

  assert (InW : set_in (rget s rsweep) W).
  assert (set_in (rget s rsweep) (set_union B W)).
  assert (set_in (rget s rsweep) ohp_ptrs).
  apply ms_objs_intro with k; trivial.
  rewrite Hed in R. 
  unfold oed in * ; unfold ost in * ; lwomega.
  rewrite <- Hobjs in H; simplset.
  tauto.
  destruct (set_union_elim H) as [InB | InW]; [tauto | trivial].
  sconjlifth 1 Hp.
  hypreplace Hp  (fas_elim InW Hp).
  sconjlifth 1 Hp.
  unfold ok_fld in Hp; sepelim Hp.
  hypreplace Hp (cont_update1 Hp (rget s rfree)).
  simplsep 7 7.
  apply free_list_intro with (rget s rfree); trivial.
  unfold ost in Hk; intro FF; rewrite FF in Hk; omega. 
  lexintro 1; exists x0.
  rewrite Plus.plus_comm; simplsep 4 4.
  sconjlift 1; sconjlifth 1 Hp.
  eapply sconj_mono; intros.
  apply Hp.

  unfold bobj_hdrs in * .
  eapply fas_impl1; intros.
  apply H.
  replace (if blt_nat a (rget s rsweep + 8) then 0 else 1)
    with (if blt_nat a (rget s rsweep) then 0 else 1).
  apply H1.
  assert (set_in a ohp_ptrs).
  simplset.
  destruct (ms_objs_elim H2) as [X1 [kk Hl]].
  simplif; simplif; try reflexivity.
  elimtype False; omega.
  assert (rget s rsweep <> a).
  intro FF; apply NinB; rewrite FF; tauto.
  elimtype False; omega.

  sconjlift 1.
  unfold obj_hdrs in * ; apply fas_intro; trivial.
  hypreplace H (fas_elim InW H).
  simplsep 2 2.

  split.
  split.
  destruct Pre as [h [B [W [F [G [[Hst Hed] [Hroot 
    [HG [HW [HF [[k Hk] [Hobjs [Hp Hh]]]]]]]]]]]]].
  assert (NinB : ~ set_in (rget s rsweep) B).
  intro InB; assert (hget s (hdr_addr (rget s rsweep)) = Some 1).
  sconjlifth 5 Hp.
  generalize (fas_elim InB Hp).
  simplif; [elimtype False; lwomega | clear Hp; intro Hp].
  simpllw 1 1; reflexivity.
  rewrite H in Case1; discriminate.
  assert (NinF : ~ set_in (rget s rsweep) F).
  intro FF.
  hypreplace FF (HF _ FF).
  rewrite HG in FF; destruct (ms_objs_elim FF).
  lwomega.
  assert (InW : set_in (rget s rsweep) W).
  assert (set_in (rget s rsweep) (set_union B W)).
  assert (set_in (rget s rsweep) ohp_ptrs).
  apply ms_objs_intro with k; trivial.
  rewrite Hed in R. 
  unfold oed in * ; unfold ost in * ; lwomega.
  rewrite <- Hobjs in H; simplset.
  tauto.
  destruct (set_union_elim H) as [InB | InW]; [tauto | trivial].
  sconjlifth 1 Hp.
  hypreplace Hp  (fas_elim InW Hp).
  sconjlifth 1 Hp.
  unfold ok_fld in Hp; sepelim Hp.
  lexintro 1; exists x; simplsep 2 7.
 
  destruct Pre as [h Pre]; exists h; auto.
  intros s' [Hg Hr]; split; auto; intros.
  rewrite Hr; trivial.
  destruct H as [A [B [C [D _]]]]; simplst; reflexivity.

  destruct R0; reflexivity.
Qed.


(** ******************************************************************* *)
(** gc() *)
(** ******************************************************************* *)
Theorem gc2_ok : 
  scap_iseq_ok gc_P gc2_pre gc2_guar gc2.
Proof.
  simplvc.
  intro s; simplpm; split; trivial.
  intros s' [Hg Hr]; split.
  split; trivial.
  rewrite Hr; [reflexivity | regin].
Qed.

Theorem gc_ok : 
  scap_iseq_ok gc_P gc_pre gc_guar gc.
Proof.
  simplvc.
  intros s Pre.
  simplpm.
  apply call_chain.

  split.
  apply Pre.
  intros s' [Hg Hr].
  assert (XXX : forall h, gc_pre_p s h ->
    exists h', gc2_pre_p s' h' /\ h ||- hp_eq h' ** top).
  intros.
  destruct (Hg _ H) as [B [W [[Hst Hed] [Hroot [Hobjs [Hp Hh]]]]]].
  destruct Hh as [h' [ht [J [Hh Ht]]]].
  assert (Hh' : h ||- hp_eq h' ** obj_hp ohp_ptrs W).
  apply sconj_intro with h' ht; trivial.
  unfold hp_eq; unfold hpred_true; reflexivity.
  hypreplace Hp (sconj_assoc_left (sconj_insert Hp Hh')).
  exists h'; split.
  exists B; exists W; simplst.
  repeat split; trivial.
  eapply roots_pred_next; eauto.
  intros r R; destruct R as [R |[R |[R | R]]]; subst; try reflexivity; tauto.
  simplsep 2 2.

  split.
  destruct Pre as [h Pre]; destruct (XXX _ Pre) as [h' [H Sub]].
  exists h'; auto.

  intros s'' [[Hg' Hr'] Hra].
  split; intros.
  destruct (XXX _ H) as [h' [H' Sub]].
  hypreplace H (Hg' _ H').
  exists h'; split; trivial.
  destruct H as [A1 [A2 [A3 [A4 [A5 [A6 [A7 [A8 _]]]]]]]].
  comparer r ra; simplst.
  rewrite Hra.
  rewrite Hr; [reflexivity | regin].
  rename e into e'.
  comparer r rtop; simplst.
  destruct Pre as [h Pre]; destruct (XXX _ Pre) as [h' [H Sub]].
  hypreplace H (Hg' _ H).
  destruct H as [B' [W' [_ [_ [_ [Hp' _]]]]]].
  destruct Pre as [_ [Hp _]].
  generalize Hp Hp'.
  unfold set_hp; unfold buffer_stack_set; 
    unfold buffer_array_stack; clear Hp Hp'; simplst; intros.
  sepelim Hp; sepelim Hp'.
  destruct A as [_ A].
  destruct A9 as [_ A9].
  rewrite <- A in A0; rewrite <- A9 in A10.
  simpl in * .
  destruct A0 as [A0 _].
  destruct A10 as [A10 _].
  rewrite <- A0; rewrite <- A10; simplpm.
  rewrite Hr'; [simplst | regin].
  rewrite Hr; [simplst; reflexivity | regin].
  rewrite Hr'; [simplst | regin].
  rewrite Hr; [simplst; reflexivity | regin].

  intros s s' [_ Hr].
  rewrite Hr; [simplst; reflexivity | regin].
Qed.


(** ******************************************************************* *)
(** alloc() *)
(** ******************************************************************* *)
Theorem ret_ptr_ok : 
  scap_iseq_ok gc_P ret_ptr_pre ret_ptr_guar ret_ptr.
Proof.
  simplvc.
  intros s [x Hx].
  simplpm; simplst; rewrite Hx.
  split; trivial.
  simplpm; intros; injection H0; intros; subst; trivial.
Qed.

Theorem alloc_loop_ok : 
  scap_iseq_ok gc_P alloc_loop_pre alloc_loop_guar alloc_loop.
Proof.
  simplvc.
  unfold alloc_loop_pre; unfold alloc_loop_guar; intros.
  simplst; simplpm; simplif; split; intros; firstorder.
Qed.

Theorem alloc_ok : 
  scap_iseq_ok gc_P alloc_pre alloc_guar alloc.
Proof.
  simplvc.
  intros s Pre.
  simplpm; simplst.
  simplif.
  apply call_chain.
  
  (* case 1: gc occured *)
  assert (XXX : forall h, gc_inv s h -> 
    gc_pre_p (rset (rset s rcst3 (rget s ra)) ra ALLOCLOOP) h).
  intros h H.
  destruct H as  [B [F [[Hst Hed] [Hobjs [Hp Hh]]]]].
  rewrite R in Hp; sconjlifth 1 Hp.
  destruct (free_list_elim0 Hp) as [Hp' E].
  subst; clear Hp; rename Hp' into Hp. 
  rewrite set_union_empty_right in Hobjs; subst.
  unfold mark_pre_p; simplst.
  repeat split; trivial.
  unfold set_hp in * ; simplst.
  simplsep 5 5.
  apply some_hp_obj_hp; trivial.

  (* alloc_pre s -> gc_pre s^ *)
  split; destruct Pre as [h Pre].
  exists h; apply XXX; trivial.

  (* gc_guar s^ s' -> alloc_loop_pre s' *)
  intros s' [Hg Hr].
  assert (Hs' := Hg _ (XXX _ Pre)).

  (* alloc_loop_pre *)
  assert (A : alloc_loop_pre s').
  destruct Hs' as [h' [[B [F [[Hst Hed] [Hroot [Hobjs [Hp' Hh']]]]]] Hh]].
  intro Neq0.
  sconjlifth 1 Hp'.
  destruct (free_list_remove Neq0 Hp') as [v [y [Hp InF]]].
  clear Hp'; rename Hp into Hp'.
  subst; exists y; rewrite Plus.plus_comm.
  simpllw y 3; reflexivity.

  split; trivial.
  intros s'' [Neq0 Hg'].
  destruct (A Neq0) as [x Hx]. 
  hypreplace Hg' (Hg' _ Hx).
  subst s''.
  clear Pre h Hs'.
  intros h Psi G T1 T2 Pre WFT1 WFT2 WFs.
  assert (GC := XXX _ Pre).
  assert (Hs' := Hg _ GC).
  destruct Hs' as [h' [[B [F [[Hst Hed] [Hroot [Hobjs [Hp' Hh']]]]]] Hh]].
  sconjlifth 1 Hp'.
  destruct (free_list_remove Neq0 Hp') as [v [y [Hp InF]]].
  clear Hp'; rename Hp into Hp'.
  assert (hget s' (rget s' rfree + 4) = Some y).
  rewrite Plus.plus_comm; simpllw y 4; reflexivity.
  rewrite Hx in H; injection H; intros; subst y; clear H.

  exists (heap_set (heap_set h' (rget s' rfree) v) (4 + rget s' rfree) x).
  split.

  (* gc_inv s'' *)
  exists (set_add (rget s' rfree) B); exists (set_remove (rget s' rfree) F).
  simplst; repeat split; trivial.
  rewrite <- Hobjs; setsolver.
  simplsep 4 4.

  apply heap_eq_add.
  sconjlift 1.  
  apply heap_eq_add1.
  simplsep 3 3.
  unfold some_hp; apply fas_intro1.
  unfold some_obj.
  lexintro 1; exists x.
  rewrite Plus.plus_comm.
  apply heap_add_ok.
  lexintro 2; exists v.
  sconjlift 1.
  apply heap_add_ok.
  apply (obj_hp_some_hp Hh').
  hypreplace Hx (heap_get_ptr Hx).
  rewrite Plus.plus_comm in Hx; simpl in Hx; trivial.
  intros ff FF.
  sconjlifth 5 Hp'.
  hypreplace Hp' (sconj_insert Hp' FF).
  sconjlifth 1 Hp'.
  apply cont_cont_false with (hp_of s') (rget s' rfree) ff v top; simplsep 3 7.
  apply (heap_get_ptr Hx).
  intros ff FF.
  assert (X : s' |||- rget s' rfree |-> v ** hp_eq h' ** 4 + rget s' rfree |-> x ** top).
  simplsep 3 7.
  hypreplace Hp' (heap_eq_add1 X); clear X.
  hypreplace Hp' (sconj_insert Hp' FF).
  rewrite Plus.plus_comm in Hp'.
  apply cont_cont_false with (hp_of s') (4 + rget s' rfree) ff x top; simplsep 3 7.
  intro FF.
  sconjlifth 5 Hp'.
  hypreplace Hp' (fas_elim FF (sconj_insert Hp' Hh')).
  unfold ok_obj in Hp'; unfold ok_fld in Hp'; unfold ok_val in Hp'; sepelim Hp'.
  apply cont_cont_false with (hp_of s') (rget s' rfree) x0 v top; simplsep 3 7.
  
  (* WFst s'' *)
  assert (WFst Psi G (rf_of (rset s' ra (rget s' rcst3)), h')).
  apply gc_step_WFst with (rf_of s, h); trivial.
  
  (* gc WFst *)
  split.
  intros addr [root [InR Rch]].
  generalize GC; clear GC.
  unfold gc_pre; unfold mark_pre_p; simplst.
  intros [[Hst0 Hed0] [Hp0 Hh0]].
  assert (InR' : In root (map (rget (rf_of s', h)) (s1 :: s2 :: rcst3 :: nil))).
  destruct s' as [rfs' hps'].
  simpl; unfold rget in * .
  repeat (rewrite Hr;[idtac | regin]).
  unfold rset; simpl.
  repeat (rewrite rfile_set_get_neq; [idtac | discriminate]).
  rewrite rfile_set_get_eq.
  destruct InR as [XX |[XX |[XX | XX]]]; intuition.
  discriminate.
  assert (InB : set_in root B).
  hypreplace Rch (vptr_in_ohp_ptrs (reach_vptrx Rch)).
  destruct InR' as [Z |[ Z |[ Z | Z]]]; try tauto; subst.
  apply (Hroot s1); [left | idtac ]; trivial.
  apply (Hroot s2); [right; left | idtac ]; trivial.
  apply (Hroot rcst3); [right; right; left | idtac ]; trivial.
  
  hypreplace InB (obj_hp_close (sconj_insert Hh Hh') InB Rch).
  unfold hget; simpl.
  apply (obj_hp_subheap_get Hh Hh' InB).

  clear Hp' XXX Hg GC.
  intros r InR.
  unfold rget in * ; unfold rset in * .
  destruct InR as [XX |[XX |[XX | XX]]]; try tauto; subst; simplst; simpl;
    simplrf; try discriminate; (rewrite Hr; [ reflexivity | regin]).

  (* alloc *)
  apply gc_step_WFst with 
    (rf_of (rset (rset s' ra (rget s' rcst3)) s2
              (rget (rset s' ra (rget s' rcst3)) rfree)),
     heap_set (heap_set h' (rget s' rfree) v) (4 + rget s' rfree) x); trivial.
  split.
  intuition.
  intros.
  unfold rget; unfold rset; simpl.
  destruct H0 as [XX |[XX |[XX | XX]]]; 
    try tauto; subst; simplrf; simplrf; try discriminate; auto.
  
  (* heap extension *)
  assert (Nin : heap_get h' (rget s' rfree) = None).
  destruct (heap_get_dec2 h' (rget s' rfree)) as [XX | [xx XX]]; trivial.
  sconjlifth 5 Hp'.  
  hypreplace Hp' (sconj_insert Hp' (cont_intro2 XX)).
  elimtype False; apply cont_cont_false with (hp_of s') (rget s' rfree) xx v top; simplsep 3 7.

  assert (Vptr : vptr (rget s' rfree)).
  apply in_ohp_ptrs_vptr.
  rewrite <- Hobjs; simplset.
  clear XXX Hg Hr A Neq0 GC  Hroot Hobjs InF Hst Hed WFs Hp' F Hh Pre.
  destruct H as [Phi [WFD WFR]].
  assert (Phif : Phi (rget s' rfree) = None).
  assert (Phi (rget s' rfree) = None \/ 
    exists C1, exists C2, Phi (rget s' rfree) = Some (C1,C2)).
  destruct (Phi (rget s' rfree)); [right | left]; auto.
  destruct p.
  exists c; exists c0; reflexivity.
  destruct H as [ H | [C1 [C2 H]]]; trivial.
  destruct WFD as [XX _].
  destruct (XX _ _ _ H) as [v1 [v2 [FF _]]].
  simpl in FF; rewrite FF in Nin; discriminate.

  exists (dhspec_set Phi (rget s' rfree) (false,T1) (false,T2)); split; dst.
  rewrite Plus.plus_comm; apply heap_extension_WFdh; trivial.
  clear WFT1.
  constructor.
  constructor.

  generalize WFR WFD; clear WFR WFD.
  unfold rset; unfold rget; dst; intros.
  simplrf.
  apply WFrf_rset.
  discriminate.
  right; left; reflexivity.
  apply heap_extension_WFrf; trivial.
  apply WFw_mem with (false, T1) (false, T2).
  simplbeq; reflexivity.
  constructor.
  constructor.
  unfold WFwty; simpl.
  splitb; trivial.

  (* case 2: no gc *)
  assert (forall h, gc_inv s h -> (exists x : nat, hget s (rget s rfree + 4) = Some x)).
  intros h H.
  destruct H as [B [F [[Hst Hed] [Hobjs [Hp Hh]]]]].
  sconjlifth 1 Hp.
  destruct (free_list_remove R Hp) as [v [y [H InF]]].
  assert (Y : hget s (rget s rfree + 4) = Some y).
  rewrite Plus.plus_comm; simpllw y 4; reflexivity.
  exists y; trivial.

  split.
  destruct Pre; apply (H _ H0).

  intros s' Hg; clear Pre.
  intros h Psi G T1 T2 Pre WFT1 WFT2 WFs.
  destruct (H _ Pre) as [y Hy]; clear H.
  hypreplace Hg (Hg _ Hy); subst s'.

  destruct Pre as [B [F [[Hst Hed] [Hobjs [Hp Hh]]]]].
  sconjlifth 1 Hp.
  destruct (free_list_remove R Hp) as [v [y' [Hp' InF]]].
  clear Hp; rename Hp' into Hp.
  assert (hget s (rget s rfree + 4) = Some y').
  rewrite Plus.plus_comm; simpllw y' 4; reflexivity.
  rewrite Hy in H; injection H; intros; subst y'; clear H.

  exists (heap_set (heap_set h (rget s rfree) v) (4 + rget s rfree) y).
  split.

  (* gc_inv s *)
  exists (set_add (rget s rfree) B); exists (set_remove (rget s rfree) F).
  simplst; repeat split; trivial.
  rewrite <- Hobjs; setsolver.
  simplsep 4 4.

  apply heap_eq_add.
  sconjlift 1.  
  apply heap_eq_add1.
  simplsep 3 3.
  unfold some_hp; apply fas_intro1.
  unfold some_obj.
  lexintro 1; exists y.
  rewrite Plus.plus_comm.
  apply heap_add_ok.
  lexintro 2; exists v.
  sconjlift 1.
  apply heap_add_ok.
  apply Hh.
  hypreplace Hy (heap_get_ptr Hy).
  rewrite Plus.plus_comm in Hy; simpl in Hy; trivial.
  intros ff FF.
  sconjlifth 5 Hp.
  hypreplace Hp (sconj_insert Hp FF).
  sconjlifth 1 Hp.
  apply cont_cont_false with (hp_of s) (rget s rfree) ff v top; simplsep 3 7.
  apply (heap_get_ptr Hy).
  intros ff FF.
  assert (X : s |||- rget s rfree |-> v ** hp_eq h ** 4 + rget s rfree |-> y ** top).
  simplsep 3 7.
  hypreplace Hp (heap_eq_add1 X); clear X.
  hypreplace Hp (sconj_insert Hp FF).
  rewrite Plus.plus_comm in Hp.
  apply cont_cont_false with (hp_of s) (4 + rget s rfree) ff y top; simplsep 3 7.
  intro FF.
  sconjlifth 5 Hp.
  hypreplace Hp (fas_elim FF (sconj_insert Hp Hh)).
  unfold some_obj in Hp; sepelim Hp.
  apply cont_cont_false with (hp_of s) (rget s rfree) x v top; simplsep 3 7.
  
  (* WFst s *)
  assert (WFst Psi G (rf_of (rset (rset s rcst3 (rget s ra)) ra (rget s ra)), h)).
  apply gc_step_WFst with (rf_of s, h); trivial.
  
  (* gc WFst *)
  split.
  intros; auto.
  intros r InR.
  unfold rget in * ; unfold rset in * .
  destruct InR as [XX |[XX |[XX | XX]]]; try tauto; subst; simplst; simpl;
    simplrf; try discriminate; reflexivity.

  (* alloc *)
  apply gc_step_WFst with 
    (rf_of (rset (rset (rset s rcst3 (rget s ra)) ra (rget s ra)) s2
      (rget s rfree)),
     heap_set (heap_set h (rget s rfree) v) (4 + rget s rfree) y); trivial.
  split.
  intuition.
  intros.
  unfold rget; unfold rset; simpl.
  destruct H0 as [XX |[XX |[XX | XX]]]; 
    try tauto; subst; simplrf; simplrf; try discriminate; auto.
  
  (* heap extension *)
  assert (Nin : heap_get h (rget s rfree) = None).
  destruct (heap_get_dec2 h (rget s rfree)) as [XX | [xx XX]]; trivial.
  sconjlifth 5 Hp.  
  hypreplace Hp (sconj_insert Hp (cont_intro2 XX)).
  elimtype False; apply cont_cont_false with (hp_of s) (rget s rfree) xx v top; simplsep 3 7.

  assert (Vptr : vptr (rget s rfree)).
  apply in_ohp_ptrs_vptr.
  rewrite <- Hobjs; simplset.
  clear Hobjs InF Hst Hed WFs Hp F Hh.
  destruct H as [Phi [WFD WFR]].
  assert (Phif : Phi (rget s rfree) = None).
  assert (Phi (rget s rfree) = None \/ 
    exists C1, exists C2, Phi (rget s rfree) = Some (C1,C2)).
  destruct (Phi (rget s rfree)); [right | left]; auto.
  destruct p.
  exists c; exists c0; reflexivity.
  destruct H as [ H | [C1 [C2 H]]]; trivial.
  destruct WFD as [XX _].
  destruct (XX _ _ _ H) as [v1 [v2 [FF _]]].
  simpl in FF; rewrite FF in Nin; discriminate.

  exists (dhspec_set Phi (rget s rfree) (false,T1) (false,T2)); split; dst.
  rewrite Plus.plus_comm; apply heap_extension_WFdh; trivial.
  clear WFT1.
  constructor.
  constructor.

  generalize WFR WFD; clear WFR WFD.
  unfold rset; unfold rget; dst; intros.
  simplrf.
  apply WFrf_rset.
  discriminate.
  right; left; reflexivity.
  apply heap_extension_WFrf; trivial.
  apply WFw_mem with (false, T1) (false, T2).
  simplbeq; reflexivity.
  constructor.
  constructor.
  unfold WFwty; simpl.
  splitb; trivial.

  intros s s' [_ Hr].
  rewrite <- Hr; [reflexivity | regin].
Qed.

