(** link all together *)

(** well-formed code heap with chase() and alloc() *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Import chaseproof.
Require Import gcproof.
Require Import scap2.


(** ******************************************************************* *)
(** globle code heap sepcification *)
(** ******************************************************************* *)
Definition P :=
  (RETURN,,(isp2 return_pre return_guar)) ::: (
  (RETURN1,,(isp2 return1_pre return1_guar)) ::: (
  (LOOP,,(isp2 loop_pre loop_guar)) ::: (

  (STACKEMPTY,,(isp2 is_empty_pre is_empty_guar)) ::: (
  (STACKEMPTYT,,(isp2 is_emptyT_pre is_emptyT_guar)) ::: (
  (STACKPOP,,(isp2 pop_pre pop_guar)) ::: (
  (STACKPUSH,,(isp2 push_pre push_guar)) ::: (

  (MARKFIELD,,(isp2 mark_field_pre mark_field_guar)) ::: (

  (MARK,,(isp2 mark_pre mark_guar)) ::: (
  (MARK1,,(isp2 mark1_pre mark1_guar)) ::: (
  (MARK2,,(isp2 mark2_pre mark2_guar)) ::: (
  (MARKLOOP,,(isp2 mark_loop_pre mark_loop_guar)) ::: (
  (MARKLOOPBODY,,(isp2 mark_loop_body_pre mark_loop_body_guar)) ::: (
  (MARKFIRST,,(isp2 mark_first_pre mark_first_guar)) ::: (
  (MARKSECOND,,(isp2 mark_second_pre mark_second_guar)) ::: (

  (SWEEP,,(isp2 sweep_pre sweep_guar)) ::: (
  (SWEEPLOOP,,(isp2 sweep_loop_pre sweep_loop_guar)) ::: (
  (SWEEPADD,,(isp2 sweep_add_pre sweep_add_guar)) ::: (
  (SWEEPNEXT,,(isp2 sweep_next_pre sweep_next_guar)) ::: (

  (GC,,(isp2 gc_pre gc_guar)) ::: (
  (GC2,,(isp2 gc2_pre gc2_guar)) ::: (

  (ALLOC,,(isp2 alloc_pre alloc_guar)) ::: (
  (ALLOCLOOP,,(isp2 alloc_loop_pre alloc_loop_guar)) ::: (
  (ALLOCRETPTR,,(isp2 ret_ptr_pre ret_ptr_guar)) ::: (

  (INIT,,(isp1 init_G)) ::: (
  (CHASE,,(isp1 chase_G)) ::: (
  (WRITE,,(isp1 write_G)) ::: (
  (RET,,(isp1 ret_G)) ::: (
  (ALLOC,,(isp1 alloc_G)) :::
  nilT _ )))))))))))))))))))))))))))).


(** ******************************************************************* *)
(** globle code heap *)
(** ******************************************************************* *)
Definition CH l :=
  if (beq_nat l RETURN) then Some returnc else
  if (beq_nat l RETURN1) then Some return1 else
  if (beq_nat l LOOP) then Some loop else 

  if (beq_nat l STACKEMPTY) then Some is_empty else
  if (beq_nat l STACKEMPTYT) then Some is_emptyT else
  if (beq_nat l STACKPOP) then Some pop else
  if (beq_nat l STACKPUSH) then Some push else

  if (beq_nat l MARKFIELD) then Some mark_field else
  if (beq_nat l MARK) then Some mark else
  if (beq_nat l MARK1) then Some mark1 else
  if (beq_nat l MARK2) then Some mark2 else
  if (beq_nat l MARKLOOP) then Some mark_loop else
  if (beq_nat l MARKLOOPBODY) then Some mark_loop_body else
  if (beq_nat l MARKFIRST) then Some mark_first else
  if (beq_nat l MARKSECOND) then Some mark_second else

  if (beq_nat l SWEEP) then Some sweep else
  if (beq_nat l SWEEPLOOP) then Some sweep_loop else
  if (beq_nat l SWEEPADD) then Some sweep_add else
  if (beq_nat l SWEEPNEXT) then Some sweep_next else

  if (beq_nat l GC) then Some gc else
  if (beq_nat l GC2) then Some gc2 else

  if (beq_nat l ALLOC) then Some alloc else
  if (beq_nat l ALLOCLOOP) then Some alloc_loop else
  if (beq_nat l ALLOCRETPTR) then Some ret_ptr else

  if (beq_nat l INIT) then Some init else
  if (beq_nat l CHASE) then Some chase else
  if (beq_nat l WRITE) then Some write else
  if (beq_nat l RET) then Some ret else
  None.


(** ******************************************************************* *)
(** alloc specification ok for tal and scap *)
(** ******************************************************************* *)
Lemma alloc_tal_scap_wk : 
  (lift (interp (isp1 alloc_G)) P) =>> (lift (interp (isp2 alloc_pre alloc_guar)) P).
Proof.
  unfold lift.
  unfold interp; unfold interp1; unfold interp2.
  intros P' [rf hp] [Hsub [ha [hb [J [Hwf Hgc]]]]].
  split; trivial.
  split.
  unfold alloc_pre.
  exists ha; trivial.
  exists 0; simpl.
  intros s' Hg.
  assert (WF1 : WFwty tint).
  unfold WFwty; reflexivity.
  assert (WF2 : WFwty tlist).
  unfold WFwty; reflexivity.
  destruct (Hg _ _ _ _ _ Hgc WF1 WF2 Hwf) as [h' [Hgc' Hwf']].
  assert (H := Hwf').
  destruct H as [Phi [[Hdh Hwfd] [Hrf Hwfr]]].
  assert (regty_get (regty_set alloc_G s2 [<tint ^^ false, tlist ^^ false>])
            ra = Some (tcode 
	      [{ (s1, tlist), (s2, [< tint ^^ false, tlist ^^ false >]), (zero, tint) }])).
  reflexivity.
  hypreplace H (Hrf _ _ H).
  inversion_clear H.

  exists [{ (s1, tlist), (s2, [< tint ^^ false, tlist ^^ false >]), (zero, tint) }].
  split; trivial.
  destruct s' as [rf' hp'].
  unfold interp1.
  assert (H := Hgc').
  destruct H as [B [F [_ [_ [Hp _]]]]].
  destruct Hp as [h0 [hb' [J' [Eq _]]]].
  unfold hp_eq in Eq; unfold hpred_true in Eq; subst h0.
  exists h'; exists hb'.
  split; trivial.
  split; trivial.
  eapply WFst_regty_sub; eauto.
  clear Hwf Hwf' Hrf Hwfr H0 Hdh.
  simpl.
  constructor.
  intros r R.
  destruct R as [R | [ R | [ R | R]]]; try tauto; subst; reflexivity.
  split; trivial.
  split; trivial.
  intros r R.
  destruct R as [R | [ R | [ R | R]]]; try tauto; subst; intuition.
  split; trivial.
  simpl; intuition; discriminate.
  right; left; reflexivity.

  inversion_clear H1.
Qed.

Opaque return_pre return_guar
  return1_pre return1_guar
  loop_pre loop_guar
  is_empty_pre is_empty_guar
  is_emptyT_pre is_emptyT_guar
  pop_pre pop_guar
  push_pre push_guar
  mark_field_pre mark_field_guar
  mark_pre mark_guar
  mark1_pre mark1_guar
  mark2_pre mark2_guar
  mark_loop_pre mark_loop_guar
  mark_loop_body_pre mark_loop_body_guar
  mark_first_pre mark_first_guar
  mark_second_pre mark_second_guar
  sweep_pre sweep_guar
  sweep_loop_pre sweep_loop_guar
  sweep_add_pre sweep_add_guar
  sweep_next_pre sweep_next_guar
  gc_pre gc_guar
  gc2_pre gc2_guar
  alloc_pre alloc_guar
  alloc_loop_pre alloc_loop_guar
  init_G chase_G write_G ret_G alloc_G

  returnc return1 loop 
  is_empty is_emptyT pop push
  mark_field mark mark1 mark2 mark_loop mark_loop_body
  mark_first mark_second
  sweep sweep_loop sweep_add sweep_next
  gc gc2
  alloc alloc_loop ret_ptr
  init chase write ret.


(** ******************************************************************* *)
(** locap well-formed code heap *)
(** ******************************************************************* *)
Lemma chase_P_subset_P : forall f T, 
  chspec_in chase_P f T -> chspec_in P f T.
Proof.
  apply dom_get_subset_chspec_subset.
  intros l H x.
  simpl in H.
  Opaque chspec_get.
  destruct H as [H |[ H |[ H |[ H |[ H | H]]]]]; try tauto; subst.
  Transparent chspec_get.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; intuition.
Qed.

Lemma gc_P_subset_P : forall f T, 
  chspec_in gc_P f T -> chspec_in P f T.
Proof.
  apply dom_get_subset_chspec_subset.
  intros l H x.
  simpl in H.
  Opaque chspec_get.
  destruct H as [H |[ H |[ H |[ H |[ H | 
    [H |[ H |[ H |[ H |[ H | 
      [H |[ H |[ H |[ H |[ H | 
	[H |[ H |[ H |[ H |[ H | 
	  [H |[ H |[ H |[ H | H ]]]]]]]]]]]]]]]]]]]]]]]]; subst.
  Transparent chspec_get.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; trivial.
  simpl; intuition.
  simpl; trivial.
  simpl; trivial.
  tauto.
Qed.

Ltac gc2locap ok :=
  match goal with
    | [ B : InT _ _ |-_] =>
      simpl in B; destruct B; 
	[subst; 
	  apply chspec_subset_iseq_ok with gc_P;
	    [apply gc_P_subset_P | apply (scap_cap_sound_iseq ok)]
	  | tauto]
  end.

Ltac tal2locap ok :=
  match goal with
    | [ B : InT _ _ |-_] =>
      simpl in B; destruct B; 
	[subst; 
	  apply chspec_subset_iseq_ok with chase_P;
	    [apply chase_P_subset_P | apply (tal_sound_iseq ok)]
	  | tauto]
  end.

Theorem code_heap_ok : ch_ok P CH P.
Proof.
  red; intros.
  destruct (chspec_in_in_dom_in_get H) as [A B]; clear H.
  simpl in A.
  destruct A as [H |[ H |[ H |[ H |[ H | 
    [H |[ H |[ H |[ H |[ H | 
      [H |[ H |[ H |[ H |[ H | 
	[H |[ H |[ H |[ H |[ H | 
	  [H |[ H |[ H |[ H |[ H | 
	    [H |[ H |[ H |[ H | H 
	    ]]]]]]]]]]]]]]]]]]]]]]]]]]]]]; 
  try tauto; subst; simpl.

  gc2locap return_ok.
  gc2locap return1_ok.
  gc2locap loop_ok.
  gc2locap is_empty_ok.
  gc2locap is_emptyT_ok.
  gc2locap pop_ok.
  gc2locap push_ok.
  gc2locap mark_field_ok.
  gc2locap mark_ok.
  gc2locap mark1_ok.
  gc2locap mark2_ok.
  gc2locap mark_loop_ok.
  gc2locap mark_loop_body_ok.
  gc2locap mark_first_ok.
  gc2locap mark_second_ok.
  gc2locap sweep_ok.
  gc2locap sweep_loop_ok.
  gc2locap sweep_add_ok.
  gc2locap sweep_next_ok.
  gc2locap gc_ok.
  gc2locap gc2_ok.

  simpl in B.
  assert (iseq_ok (lift (interp (isp2 alloc_pre alloc_guar)) P) alloc).
  apply chspec_subset_iseq_ok with gc_P;
    [apply gc_P_subset_P | apply (scap_cap_sound_iseq alloc_ok)].
  destruct B as [B | [ B | B]]; try tauto; subst; trivial.
  eapply iseq_ok_weakening; eauto.
  apply alloc_tal_scap_wk.

  gc2locap alloc_loop_ok.
  gc2locap ret_ptr_ok.

  tal2locap init_ok.
  tal2locap chase_ok.
  tal2locap write_ok.
  tal2locap ret_ok.

  simpl in B.
  assert (iseq_ok (lift (interp (isp2 alloc_pre alloc_guar)) P) alloc).
  apply chspec_subset_iseq_ok with gc_P;
    [apply gc_P_subset_P | apply (scap_cap_sound_iseq alloc_ok)].
  destruct B as [B | [ B | B]]; try tauto; subst; trivial.
  eapply iseq_ok_weakening; eauto.
  apply alloc_tal_scap_wk.
Qed.
